import hashlib
import logging
import os
import shutil
import stat
import sys
import tempfile
import uuid
from collections.abc import Iterator, Iterable
from contextlib import contextmanager
from io import BytesIO
from typing import IO, Any, Callable, Optional, Protocol, Union

from toil.lib.memoize import memoize
from toil.lib.misc import StrPath

logger = logging.getLogger(__name__)


@memoize
def get_toil_home() -> str:
    """
    Get the Toil home directory for storing configuration and global state.

    Raises an error if it does not exist and cannot be created. Safe to run
    simultaneously in multiple processes.
    """
   
    # TODO: should this use an XDG config directory or ~/.config to not clutter the
    # base home directory?
    toil_home_dir = os.path.join(os.path.expanduser("~"), ".toil")

    dir_path = try_path(toil_home_dir)
    if dir_path is None:
        raise RuntimeError(
            f"Cannot create or access Toil configuration directory {toil_home_dir}"
        )
    return dir_path

TOIL_URI_SCHEME = "toilfile:"


STANDARD_SCHEMES = ["http:", "https:", "s3:", "gs:", "ftp:"]
REMOTE_SCHEMES = STANDARD_SCHEMES + [TOIL_URI_SCHEME]
ALL_SCHEMES = REMOTE_SCHEMES + ["file:"]

def is_standard_url(filename: str) -> bool:
    """
    Return True if the given URL is a non-Toil, non-file: URL.
    """
    return is_url_with_scheme(filename, STANDARD_SCHEMES)

def is_remote_url(filename: str) -> bool:
    """
    Decide if a filename is a known, non-file kind of URL
    """
    return is_url_with_scheme(filename, REMOTE_SCHEMES)

def is_any_url(filename: str) -> bool:
    """
    Decide if a string is a URI like http:// or file://.

    Otherwise it might be a bare path.
    """
    return is_url_with_scheme(filename, ALL_SCHEMES)

def is_url_with_scheme(filename: str, schemes: list[str]) -> bool:
    """
    Return True if filename is a URL with any of the given schemes and False otherwise.
    """
    # TODO: "http:myfile.dat" is a valid filename and *not* a valid URL
    for scheme in schemes:
        if filename.startswith(scheme):
            return True
    return False

def is_toil_url(filename: str) -> bool:
    """
    Return True if a URL is a toilfile: URL.
    """
    return is_url_with_scheme(filename, [TOIL_URI_SCHEME])

def is_file_url(filename: str) -> bool:
    """
    Return True if a URL is a file: URL.

    Will return False for bare paths.
    """
    return is_url_with_scheme(filename, ["file:"])

def mkdtemp(
    suffix: Optional[str] = None,
    prefix: Optional[str] = None,
    dir: Optional[StrPath] = None,
) -> str:
    """
    Make a temporary directory like tempfile.mkdtemp, but with relaxed permissions.

    The permissions on the directory will be 711 instead of 700, allowing the
    group and all other users to traverse the directory. This is necessary if
    the directory is on NFS and the Docker daemon would like to mount it or a
    file inside it into a container, because on NFS even the Docker daemon
    appears bound by the file permissions.

    See <https://github.com/DataBiosphere/toil/issues/4644>, and
    <https://stackoverflow.com/a/67928880> which talks about a similar problem
    but in the context of user namespaces.
    """
    # Make the directory
    result = tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dir)
    # Grant all the permissions: full control for user, and execute for group and other
    os.chmod(
        result, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    )
    # Return the path created
    return result


def robust_rmtree(path: Union[str, bytes]) -> None:
    """
    Robustly tries to delete paths.

    Continues silently if the path to be removed is already gone, or if it
    goes away while this function is executing.

    May raise an error if a path changes between file and directory while the
    function is executing, or if a permission error is encountered.
    """
    # TODO: only allow str or bytes as an input
    if not isinstance(path, bytes):
        # Internally we must work in bytes, in case we find an undecodeable
        # filename.
        path = path.encode("utf-8")

    if not os.path.exists(path):
        # Nothing to do!
        return

    if not os.path.islink(path) and os.path.isdir(path):
        # It is or has been a directory

        try:
            children = os.listdir(path)
        except FileNotFoundError:
            # Directory went away
            return
        except OSError as exc:
            if exc.errno == 16:
                # 'Device or resource busy'
                return
            raise

        # We assume the directory going away while we have it open won't upset
        # the listdir iterator.
        for child in children:
            # Get the path for each child item in the directory
            child_path = os.path.join(path, child)

            # Remove it if still present
            robust_rmtree(child_path)

        try:
            # Actually remove the directory once the children are gone
            shutil.rmtree(path)
        except FileNotFoundError:
            # Directory went away
            return
        except OSError as exc:
            if exc.errno == 16:
                # 'Device or resource busy'
                return
            raise

    else:
        # It is not or was not a directory.
        try:
            # Unlink it as a normal file
            os.unlink(path)
        except FileNotFoundError:
            # File went away
            return
        except OSError as exc:
            if exc.errno == 16:
                # 'Device or resource busy'
                return
            raise


def atomic_tmp_file(final_path: StrPath) -> str:
    """Return a tmp file name to use with atomic_install.  This will be in the
    same directory as final_path. The temporary file will have the same extension
    as finalPath.  It the final path is in /dev (/dev/null, /dev/stdout), it is
    returned unchanged and atomic_tmp_install will do nothing."""
    final_dir = os.path.dirname(os.path.normpath(final_path))  # can be empty
    if final_dir == "/dev":
        return str(final_path)
    final_basename = os.path.basename(final_path)
    final_ext = os.path.splitext(final_path)[1]
    base_name = f"{final_basename}.{str(uuid.uuid4())}.tmp{final_ext}"
    return os.path.join(final_dir, base_name)


def atomic_install(tmp_path: StrPath, final_path: StrPath) -> None:
    """atomic install of tmp_path as final_path"""
    if os.path.dirname(os.path.normpath(final_path)) != "/dev":
        os.rename(tmp_path, final_path)


@contextmanager
def AtomicFileCreate(final_path: StrPath, keep: bool = False) -> Iterator[str]:
    """Context manager to create a temporary file.  Entering returns path to
    the temporary file in the same directory as finalPath.  If the code in
    context succeeds, the file renamed to its actual name.  If an error
    occurs, the file is not installed and is removed unless keep is specified.
    """
    tmp_path = atomic_tmp_file(final_path)
    try:
        yield tmp_path
        atomic_install(tmp_path, final_path)
    except Exception:
        if not keep:
            try:
                os.unlink(tmp_path)
            except Exception:
                pass
        raise


def atomic_copy(
    src_path: str, dest_path: str, executable: Optional[bool] = None
) -> None:
    """Copy a file using posix atomic creations semantics."""
    if executable is None:
        executable = os.stat(src_path).st_mode & stat.S_IXUSR != 0
    with AtomicFileCreate(dest_path) as dest_path_tmp:
        shutil.copyfile(src_path, dest_path_tmp)
        if executable:
            os.chmod(dest_path_tmp, os.stat(dest_path_tmp).st_mode | stat.S_IXUSR)


def atomic_copyobj(
    src_fh: BytesIO, dest_path: str, length: int = 16384, executable: bool = False
) -> None:
    """Copy an open file using posix atomic creations semantics."""
    with AtomicFileCreate(dest_path) as dest_path_tmp:
        with open(dest_path_tmp, "wb") as dest_path_fh:
            shutil.copyfileobj(src_fh, dest_path_fh, length=length)
        if executable:
            os.chmod(dest_path_tmp, os.stat(dest_path_tmp).st_mode | stat.S_IXUSR)


def make_public_dir(in_directory: str, suggested_name: Optional[str] = None) -> str:
    """
    Make a publicly-accessible directory in the given directory.

    :param suggested_name: Use this directory name first if possible.

    Try to make a random directory name with length 4 that doesn't exist, with the given prefix.
    Otherwise, try length 5, length 6, etc, up to a max of 32 (len of uuid4 with dashes replaced).
    This function's purpose is mostly to avoid having long file names when generating directories.
    If somehow this fails, which should be incredibly unlikely, default to a normal uuid4, which was
    our old default.
    """
    if suggested_name is not None:
        generated_dir_path1 = os.path.join(in_directory, suggested_name)
        try:
            os.mkdir(generated_dir_path1)
            os.chmod(generated_dir_path1, 0o777)
            return generated_dir_path1
        except FileExistsError:
            pass
    for i in range(
        4, 32 + 1
    ):  # make random uuids and truncate to lengths starting at 4 and working up to max 32
        for _ in range(10):  # make 10 attempts for each length
            truncated_uuid = str(uuid.uuid4()).replace("-", "")[:i]
            generated_dir_path2 = os.path.join(in_directory, truncated_uuid)
            try:
                os.mkdir(generated_dir_path2)
                os.chmod(generated_dir_path2, 0o777)
                return generated_dir_path2
            except FileExistsError:
                pass
    this_should_never_happen: str = os.path.join(in_directory, str(uuid.uuid4()))
    os.mkdir(this_should_never_happen)
    os.chmod(this_should_never_happen, 0o777)
    return this_should_never_happen


def try_path(path: str, min_size: int = 100 * 1024 * 1024) -> Optional[str]:
    """
    Try to use the given path. Return it if it exists or can be made,
    and we can make things within it, or None otherwise.

    :param min_size: Reject paths on filesystems smaller than this many bytes.
    """

    try:
        os.makedirs(path, exist_ok=True)
    except OSError:
        # Maybe we lack permissions
        return None

    if not os.path.exists(path):
        # We didn't manage to make it
        return None

    if not os.access(path, os.W_OK):
        # It doesn't look writable
        return None

    try:
        stats = os.statvfs(path)
    except OSError:
        # Maybe we lack permissions
        return None

    # Is the filesystem big enough?
    # We need to look at the FS size and not the free space so we don't change
    # over to a different filesystem when this one fills up.
    fs_size = stats.f_frsize * stats.f_blocks
    if fs_size < min_size:
        # Too small
        return None

    return path


class WriteWatchingStream:
    """
    A stream wrapping class that calls any functions passed to onWrite() with the number of bytes written for every write.

    Not seekable.
    """

    def __init__(self, backingStream: IO[Any]) -> None:
        """
        Wrap the given backing stream.
        """

        self.backingStream = backingStream
        # We have no write listeners yet
        self.writeListeners: list[Callable[[int], None]] = []

    def onWrite(self, listener: Callable[[int], None]) -> None:
        """
        Call the given listener with the number of bytes written on every write.
        """

        self.writeListeners.append(listener)

    # Implement the file API from https://docs.python.org/2.4/lib/bltin-file-objects.html

    def write(self, data: bytes) -> None:
        """
        Write the given data to the file.
        """

        # Do the write
        self.backingStream.write(data)

        for listener in self.writeListeners:
            # Send out notifications
            listener(len(data))

    def writelines(self, datas: Iterable[bytes]) -> None:
        """
        Write each string from the given iterable, without newlines.
        """

        for data in datas:
            self.write(data)

    def flush(self) -> None:
        """
        Flush the backing stream.
        """

        self.backingStream.flush()

    def close(self) -> None:
        """
        Close the backing stream.
        """

        self.backingStream.close()

class ReadableFileObj(Protocol):
    """
    Protocol that is more specific than what file_digest takes as an argument.
    Also guarantees a read() method.
    Would extend the protocol from Typeshed for hashlib but those are only
    declared for 3.11+.
    """
    def readinto(self, buf: bytearray, /) -> int: ...
    def readable(self) -> bool: ...
    def read(self, number: int) -> bytes: ...

# hashlib._Hash seems to not appear at runtime
def file_digest(f: ReadableFileObj, alg_name: str) -> "hashlib._Hash":
    """
    Polyfilled hashlib.file_digest that works on Python <3.11.
    """
    if sys.version_info >= (3, 11):
        return hashlib.file_digest(f, alg_name)
    BUFFER_SIZE = 1024 * 1024
    hasher = hashlib.new(alg_name)
    buffer = f.read(BUFFER_SIZE)
    while buffer:
        hasher.update(buffer)
        buffer = f.read(BUFFER_SIZE)
    return hasher





# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Terminates the specified cluster and associated resources."""
import logging

from toil.common import parser_with_common_options
from toil.provisioners import cluster_factory
from toil.statsAndLogging import set_logging_from_options

logger = logging.getLogger(__name__)


def main() -> None:
    parser = parser_with_common_options(
        provisioner_options=True, jobstore_option=False, prog="toil destroy-cluster"
    )
    options = parser.parse_args()
    set_logging_from_options(options)

    logger.info("Destroying cluster %s", options.clusterName)

    cluster = cluster_factory(
        provisioner=options.provisioner,
        clusterName=options.clusterName,
        zone=options.zone,
    )
    cluster.destroyCluster()

    logger.info("Cluster %s is now gone.", options.clusterName)



# Toil documentation build configuration file, created by
# sphinx-quickstart on Tue Aug 25 12:37:16 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.
import sys
import os
import inspect
import re
from datetime import datetime
import toil.version
import time

# This makes the modules located in docs/vendor/sphinxcontrib available to import
sphinxPath = os.path.abspath(
    os.path.join(os.path.pardir, os.path.dirname("docs/vendor/sphinxcontrib/"))
)
sys.path.append(sphinxPath)
import fulltoc


def fetch_parent_dir(filepath: str, n: int = 1):
    """
    Returns a parent directory, n places above the input filepath.

    Equivalent to something like: '/home/user/dir'.split('/')[-2] if n=2.
    """
    filepath = os.path.realpath(filepath)
    for i in range(n):
        filepath = os.path.dirname(filepath)
    return os.path.basename(filepath)


path_to_dir = os.path.dirname(os.path.abspath(__file__))

# Example of toil.version.__file__ on sphinx:
# /home/docs/checkouts/readthedocs.org/user_builds/toil/envs/3.13.0/local/lib/python2.7/site-packages/toil-3.13.0a1-py2.7.egg/toil/version.pyc
envPath = os.path.abspath(toil.version.__file__)

# Example of __file__ on sphinx:
# /home/docs/checkouts/readthedocs.org/user_builds/toil/checkouts/3.13.0/docs/conf.py
wdPath_version = fetch_parent_dir(__file__, 2)
# Expected output: 3.13.0

assert (
    wdPath_version in envPath
), "Another Toil installation seems to have precedence over this working directory."
toilVersion = toil.version.baseVersion

# -- General configuration ------------------------------------------------
# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.autosummary",
    "autoapi.extension",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.autoprogram",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "fulltoc",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "cwltool": ("https://cwltool.readthedocs.io/en/latest/", None),
}

# Link definitions available everywhere so we don't need to keep repeating ourselves.
rst_epilog = """
.. _Common Workflow Language: http://www.commonwl.org/
"""


def skip(app, what, name, obj, skip, options):
    """
    Decide what to automatically generate documentation for.
    """

    # See
    # <https://sphinx-autoapi.readthedocs.io/en/latest/reference/config.html#event-autoapi-skip-member>

    # TODO: It's not clear that the contidions used here are a good idea. Why
    # are they like this?

    # Always document __init__
    return name != "__init__" and (
        skip
        # But don't document classes
        or inspect.isclass(obj)
        # Or things starting with _ that aren't modules
        or (
            name.startswith("_")
            and not inspect.ismodule(obj)
        )
    )


def setup(app):
    app.connect("autodoc-skip-member", skip)


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.  Specify multiple suffix as list of string:
# Example: source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Toil"
build_date = datetime.utcfromtimestamp(
    int(os.environ.get("SOURCE_DATE_EPOCH", time.time()))
)
copyright = f"2015 – {build_date.year} UCSC Computational Genomics Lab"
author = "UCSC Computational Genomics Lab"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = re.split("[A-Za-z]", toilVersion)[0]
# The full version, including alpha/beta/rc tags.
release = toilVersion

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, 2do and 2doList produce output, else they produce nothing.
todo_include_todos = True

# Include doc string for __init__ method in the documentation
autoclass_content = "class"
autodoc_member_order = "bysource"

autoapi_dirs = ["../src/toil"]
autodoc_typehints = "description"
autoapi_keep_files = True
autoapi_ignore = ["*.pyi", "*/test/cwl/spec*/*.py", "*/fake_mpi_run.py", "*/tutorial_*.py", "*/example_*.py", "*/mkFile.py", "*/debugWorkflow.py",
                  "*/test/wdl/wdl-conformance-tests/*"]
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
]
always_document_param_types = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "alabaster"
html_theme_options = {
    "github_banner": True,
    "github_user": "BD2KGenomics",
    "github_repo": "toil",
    "caption_font_size": "24px",
}

# The name of an image file (relative to this directory) to place at the top of the sidebar.
html_logo = "_static/logo.jpeg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# Output file base name for HTML help builder.
htmlhelp_basename = "Toildoc"

# -- Options for LaTeX output ---------------------------------------------
latex_elements = {}
# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto, manual, or own class]).
latex_documents = [(master_doc, "Toil.tex", "Toil Documentation", author, "manual")]

# -- Options for manual page output ---------------------------------------
# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "toil", "Toil Documentation", [author], 1)]

# -- Options for Texinfo output -------------------------------------------
# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author, dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        project,
        "Toil Documentation",
        author,
        project,
        project,
        "Miscellaneous",
    )
]



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import mimetypes
import os
from pathlib import Path
import subprocess
import sys
from typing import Optional

from toil.test import pslow as slow
from toil.test.mesos import helloWorld

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class RegularLogTest:

    def _getFiles(self, dirpath: Path) -> list[str]:
        return [
            os.path.join(dirpath, f)
            for f in os.listdir(dirpath)
            if os.path.isfile(os.path.join(dirpath, f))
        ]

    def _assertFileTypeExists(
        self, dirpath: Path, extension: str, encoding: Optional[str] = None
    ) -> None:
        # an encoding of None implies no compression
        logger.info("Checking for %s file in %s", extension, dirpath)
        onlyFiles = self._getFiles(dirpath)
        logger.info("Found: %s", str(os.listdir(dirpath)))
        onlyLogs = [f for f in onlyFiles if f.endswith(extension)]
        logger.info("Found matching: %s", str(onlyLogs))
        assert onlyLogs

        if encoding is not None:
            for log in onlyLogs:
                with open(log, "rb") as f:
                    logger.info(
                        "Checking for encoding %s on file %s", str(encoding), log
                    )
                    if encoding == "gzip":
                        # Check for gzip magic header '\x1f\x8b'
                        assert f.read().startswith(b"\x1f\x8b")
                    else:
                        mime = mimetypes.guess_type(log)
                        assert mime[1] == encoding

    @slow
    def testLogToMaster(self) -> None:
        toilOutput = subprocess.check_output(
            [
                sys.executable,
                "-m",
                helloWorld.__name__,
                "./toilTest",
                "--clean=always",
                "--logLevel=info",
            ],
            stderr=subprocess.STDOUT,
        )
        assert helloWorld.childMessage in toilOutput.decode("utf-8")

    def testWriteLogs(self, tmp_path: Path) -> None:
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                helloWorld.__name__,
                "./toilTest",
                "--clean=always",
                "--logLevel=debug",
                f"--writeLogs={tmp_path}",
            ]
        )
        self._assertFileTypeExists(tmp_path, ".log")

    @slow
    def testWriteGzipLogs(self, tmp_path: Path) -> None:
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                helloWorld.__name__,
                "./toilTest",
                "--clean=always",
                "--logLevel=debug",
                f"--writeLogsGzip={tmp_path}",
            ]
        )
        self._assertFileTypeExists(tmp_path, ".log.gz", "gzip")

    @slow
    def testMultipleLogToMaster(self) -> None:
        toilOutput = subprocess.check_output(
            [
                sys.executable,
                "-m",
                helloWorld.__name__,
                "./toilTest",
                "--clean=always",
                "--logLevel=info",
            ],
            stderr=subprocess.STDOUT,
        )
        assert helloWorld.parentMessage in toilOutput.decode("utf-8")

    def testRegularLog(self) -> None:
        toilOutput = subprocess.check_output(
            [
                sys.executable,
                "-m",
                helloWorld.__name__,
                "./toilTest",
                "--clean=always",
                "--batchSystem=single_machine",
                "--logLevel=debug",
            ],
            stderr=subprocess.STDOUT,
        )
        assert "single machine batch system" in toilOutput.decode("utf-8")



import pytest


@pytest.fixture(scope="class")
def rootpath(request: pytest.FixtureRequest) -> None:
    """Records the rootpath at the class level, for use on a unittest.TestCase."""
    request.cls._rootpath = request.config._rootpath



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from toil.exceptions import FailedJobsException
from toil.job import Job, JobFunctionWrappingJob
from toil.jobStores.abstractJobStore import NoSuchFileException
from toil.test import pslow as slow

import pytest


class TestResumability:
    """
    https://github.com/BD2KGenomics/toil/issues/808
    """

    @slow
    @pytest.mark.slow
    def test(self, tmp_path: Path) -> None:
        """
        Tests that a toil workflow that fails once can be resumed without a NoSuchJobException.
        """
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.logLevel = "INFO"
        options.retryCount = 0
        root = Job.wrapJobFn(parent)
        with pytest.raises(FailedJobsException):
            # This one is intended to fail.
            Job.Runner.startToil(root, options)

        # Resume the workflow. Unfortunately, we have to check for
        # this bug using the logging output, since although the
        # NoSuchJobException causes the worker to fail, the batch
        # system code notices that the job has been deleted despite
        # the failure and avoids the failure.
        options.restart = True
        tempDir = tmp_path / "tempdir"
        tempDir.mkdir()
        options.logFile = str(tempDir / "log.txt")
        Job.Runner.startToil(root, options)
        with open(options.logFile) as f:
            logString = f.read()
            # We are looking for e.g. "Batch system is reporting that
            # the jobGraph with batch system ID: 1 and jobGraph
            # store ID: n/t/jobwbijqL failed with exit value 1"
            assert "failed with exit value" not in logString

    def test_chaining(self, tmp_path: Path) -> None:
        """
        Tests that a job which is chained to and fails can resume and succeed.
        """

        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.logLevel = "DEBUG"
        options.retryCount = 0
        tempDir = tmp_path / "tempdir"
        tempDir.mkdir()
        options.logFile = str(tempDir / "log.txt")

        root = Job.wrapJobFn(chaining_parent)

        with pytest.raises(FailedJobsException):
            # This one is intended to fail.
            Job.Runner.startToil(root, options)

        with open(options.logFile) as f:
            log_content = f.read()
            # Make sure we actually did do chaining
            assert "Chaining from" in log_content

        # Because of the chaining, the problem we are looking for is the job
        # with the root ID not being able to load the body of a job with a
        # different ID. That doesn't look like a job deleted despite failure.
        options.restart = True
        Job.Runner.startToil(root, options)


def parent(job: Job) -> None:
    """
    Set up a bunch of dummy child jobs, and a bad job that needs to be
    restarted as the follow on.
    """
    for _ in range(5):
        job.addChildJobFn(goodChild)
    job.addFollowOnJobFn(badChild)


def chaining_parent(job: Job) -> None:
    """
    Set up a failing job to chain to.
    """
    job.addFollowOnJobFn(badChild)


def goodChild(job: Job) -> None:
    """
    Does nothing.
    """
    return


def badChild(job: JobFunctionWrappingJob) -> None:
    """
    Fails the first time it's run, succeeds the second time.
    """
    try:
        with job.fileStore.jobStore.read_shared_file_stream("alreadyRun") as fileHandle:
            fileHandle.read()
    except NoSuchFileException as ex:
        with job.fileStore.jobStore.write_shared_file_stream(
            "alreadyRun", encrypted=False
        ) as fileHandle:
            fileHandle.write(b"failed once\n")
        raise RuntimeError(f"this is an expected error: {str(ex)}")



from __future__ import absolute_import
import os
import sys
import ast
import tempfile
import shutil


def enable_absolute_imports(script, script_name):
    """
    Empty modules
    >>> enable_absolute_imports('')
    'from __future__ import absolute_import\\n'

    Ignore empty lines
    >>> enable_absolute_imports('\\n')
    'from __future__ import absolute_import\\n'

    Append after initial comments, like shebangs
    >>> enable_absolute_imports('#foo\\n')
    '#foo\\nfrom __future__ import absolute_import\\n'

    Insert before regular comments
    >>> enable_absolute_imports('#foo\\nimport bar\\n')
    '#foo\\nfrom __future__ import absolute_import\\nimport bar\\n'

    Insert before non-import statements
    >>> enable_absolute_imports('if False:\\n    pass\\n')
    'from __future__ import absolute_import\\nif False:\\n    pass\\n'

    Idempotence
    >>> enable_absolute_imports('from __future__ import absolute_import\\n') is None
    True

    Other __future__ imports
    >>> enable_absolute_imports('from __future__ import print_function\\n')
    'from __future__ import absolute_import\\nfrom __future__ import print_function\\n'

    Insert before from ... immport statements
    >>> enable_absolute_imports('from blah import fasel\\n')
    'from __future__ import absolute_import\\nfrom blah import fasel\\n'

    Insert before multiple future imports
    >>> enable_absolute_imports('from __future__ import print_function\\nfrom __future__ import nested_scopes\\n')
    'from __future__ import absolute_import\\nfrom __future__ import print_function\\nfrom __future__ import nested_scopes\\n'

    Insert before wrapped multi-name future import
    >>> enable_absolute_imports('from __future__ import (print_function,\\n    nested_scopes)\\n')
    'from __future__ import absolute_import\\nfrom __future__ import (print_function,\\n    nested_scopes)\\n'

    Usually docstrings show up as attributes of other nodes but unassociated docstring become
    Expr nodes in the AST.
    >>> enable_absolute_imports("#foo\\n\\n'''bar'''\\n\\npass")
    "#foo\\n\\nfrom __future__ import absolute_import\\n'''bar'''\\n\\npass\\n"

    Unassociated multiline docstring
    >>> enable_absolute_imports("#foo\\n\\n'''bar\\n'''\\n\\npass")
    "#foo\\n\\nfrom __future__ import absolute_import\\n'''bar\\n'''\\n\\npass\\n"
    """
    tree = ast.parse(script, filename=script_name)
    lines = script.split('\n')
    while lines and lines[-1] == "":
        lines.pop()
    node = None
    for child in ast.iter_child_nodes(tree):
        if isinstance(child, ast.Import):
            node = child
            break
        elif isinstance(child, ast.ImportFrom):
            assert child.level == 0  # don't know what this means
            if child.module == '__future__':
                if any(alias.name == 'absolute_import' for alias in child.names):
                    return None
                else:
                    if node is None: node = child
            else:
                node = child
                break
    if node is None:
        if len(tree.body) == 0:
            node = ast.stmt()
            node.lineno = len(lines) + 1
        else:
            node = tree.body[0]
            # This crazy heuristic tries to handle top-level docstrings with newlines in them
            # for which lineno is the line where the docstring ends
            if isinstance(node, ast.Expr) and isinstance(node.value, ast.Str):
                node.lineno -= node.value.s.count('\n')

    line = 'from __future__ import absolute_import'
    lines.insert(node.lineno - 1, line)
    lines.append("")
    return '\n'.join(lines)


def main(root_path):
    for dir_path, dir_names, file_names in os.walk(root_path):
        for file_name in file_names:
            if file_name.endswith('.py') and file_name != 'setup.py':
                file_path = os.path.join(dir_path, file_name)
                with open(file_path) as file:
                    script = file.read()
                new_script = enable_absolute_imports(script, file_name)
                if new_script is not None:
                    temp_handle, temp_file_path = tempfile.mkstemp(prefix=file_name, dir=dir_path)
                    try:
                        with os.fdopen(temp_handle, 'w') as temp_file:
                            temp_file.write(new_script)
                    except:
                        os.unlink(temp_file_path)
                        raise
                    else:
                        shutil.copymode(file_path,temp_file_path)
                        os.rename(temp_file_path, file_path)


if __name__ == '__main__':
    main(sys.argv[1])



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import errno
import logging
import os
import pickle
import random
import re
import shutil
import stat
import time
import uuid
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from typing import IO, Literal, Optional, Union, overload
from urllib.parse import ParseResult, quote, unquote

from toil.fileStores import FileID
from toil.job import TemporaryID
from toil.jobStores.abstractJobStore import (
    AbstractJobStore,
    JobStoreExistsException,
    NoSuchFileException,
    NoSuchJobException,
    NoSuchJobStoreException,
)
from toil.lib.io import (
    AtomicFileCreate,
    atomic_copy,
    atomic_copyobj,
    mkdtemp,
    robust_rmtree,
)

logger = logging.getLogger(__name__)


class FileJobStore(AbstractJobStore):
    """
    A job store that uses a directory on a locally attached file system. To be compatible with
    distributed batch systems, that file system must be shared by all worker nodes.
    """

    # Valid chars for the creation of temporary "spray" directories.
    validDirs = "abcdefghijklmnopqrstuvwxyz0123456789"
    validDirsSet = set(validDirs)

    # What prefix should be on the per-job job directories, to distinguish them
    # from the spray directories?
    JOB_DIR_PREFIX = "instance-"

    # What prefix do we put on the per-job-name directories we sort jobs into?
    JOB_NAME_DIR_PREFIX = "kind-"

    # 10Mb RAM chunks when reading/writing files
    BUFFER_SIZE = 10485760  # 10Mb

    # When a log file is still being written, what will its name end with?
    LOG_TEMP_SUFFIX = ".new"
    # All log files start with this prefix
    LOG_PREFIX = "stats"

    def default_caching(self) -> bool:
        """
        Jobstore's preference as to whether it likes caching or doesn't care about it.
        Some jobstores benefit from caching, however on some local configurations it can be flaky.

        see https://github.com/DataBiosphere/toil/issues/4218
        """

        return False

    def __init__(self, path: str, fanOut: int = 1000) -> None:
        """
        :param path: Path to directory holding the job store
        :param fanOut: Number of items to have in a directory before making
                           subdirectories
        """
        super().__init__(path)
        self.jobStoreDir = os.path.abspath(path)
        logger.debug("Path to job store directory is '%s'.", self.jobStoreDir)

        # Directory where actual job files go, and their job-associated temp files
        self.jobsDir = os.path.join(self.jobStoreDir, "jobs")
        # Directory where stats files go
        self.statsDir = os.path.join(self.jobStoreDir, "stats")
        # Which has subdirectories for new and seen stats files
        self.stats_inbox = os.path.join(self.statsDir, "inbox")
        self.stats_archive = os.path.join(self.statsDir, "archive")
        # Directory where non-job-associated files for the file store go
        self.filesDir = os.path.join(self.jobStoreDir, "files/no-job")
        # Directory where job-associated files for the file store go.
        # Each per-job directory in here will have separate directories for
        # files to clean up and files to not clean up when the job is deleted.
        self.jobFilesDir = os.path.join(self.jobStoreDir, "files/for-job")
        # Directory where shared files go
        self.sharedFilesDir = os.path.join(self.jobStoreDir, "files/shared")

        self.fanOut = fanOut

        self.linkImports = None
        self.moveExports = None
        self.symlink_job_store_reads = None

    def __repr__(self):
        return f"FileJobStore({self.jobStoreDir})"

    def initialize(self, config):
        try:
            os.mkdir(self.jobStoreDir)
        except OSError as e:
            if e.errno == errno.EEXIST:
                raise JobStoreExistsException(self.jobStoreDir, "file")
            else:
                raise
        os.makedirs(self.jobsDir, exist_ok=True)
        os.makedirs(self.statsDir, exist_ok=True)
        os.makedirs(self.stats_inbox, exist_ok=True)
        os.makedirs(self.stats_archive, exist_ok=True)
        os.makedirs(self.filesDir, exist_ok=True)
        os.makedirs(self.jobFilesDir, exist_ok=True)
        os.makedirs(self.sharedFilesDir, exist_ok=True)
        self.linkImports = config.symlinkImports
        self.moveExports = config.moveOutputs
        self.symlink_job_store_reads = config.symlink_job_store_reads
        super().initialize(config)

    def resume(self):
        if not os.path.isdir(self.jobStoreDir):
            raise NoSuchJobStoreException(self.jobStoreDir, "file")
        super().resume()
        # TODO: Unify with initialize() configuration
        self.linkImports = self.config.symlinkImports
        self.moveExports = self.config.moveOutputs
        self.symlink_job_store_reads = self.config.symlink_job_store_reads

    def destroy(self):
        if os.path.exists(self.jobStoreDir):
            robust_rmtree(self.jobStoreDir)

    ##########################################
    # The following methods deal with creating/loading/updating/writing/checking for the
    # existence of jobs
    ##########################################

    def assign_job_id(self, job_description):
        # Get the job's name. We want to group jobs with the same name together.
        # This will be e.g. the function name for wrapped-function jobs.
        # Make sure to render it filename-safe
        usefulFilename = self._make_string_filename_safe(job_description.jobName)

        # Make a unique temp directory under a directory for this job name,
        # possibly sprayed across multiple levels of subdirectories.
        absJobDir = mkdtemp(
            prefix=self.JOB_DIR_PREFIX,
            dir=self._get_arbitrary_jobs_dir_for_name(usefulFilename),
        )

        job_description.jobStoreID = self._get_job_id_from_dir(absJobDir)

    def create_job(self, job_description):
        if hasattr(self, "_batchedUpdates") and self._batchedUpdates is not None:
            # Save it later
            self._batchedUpdates.append(job_description)
        else:
            # Save it now
            self.update_job(job_description)
        return job_description

    @contextmanager
    def batch(self):
        self._batchedUpdates = []
        yield
        for jobDescription in self._batchedUpdates:
            self.update_job(jobDescription)
        self._batchedUpdates = None

    def _wait_for_exists(self, jobStoreID, maxTries=35, sleepTime=1):
        """
        Spin-wait and block for a job to appear before returning
        False if it does not.
        """
        return self._wait_for_file(
            self._get_job_file_name(jobStoreID), maxTries=maxTries, sleepTime=sleepTime
        )

    def _wait_for_file(self, fileName, maxTries=35, sleepTime=1):
        """
        Spin-wait and block for a file or directory to appear before returning
        False if it does not.

        The total max wait time is maxTries * sleepTime. The current default is
        tuned to match Linux NFS defaults where the client's cache of the directory
        listing on the server is supposed to become coherent within 30 sec.
        Delayes beyond that would probably indicate a pathologically slow file system
        that just should not be used for the jobStore.

        The warning will be sent to the log only on the first retry.

        In practice, the need for retries happens rarely, but it does happen
        over the course of large workflows with a jobStore on a busy NFS.
        """
        for iTry in range(1, maxTries + 1):
            if os.path.exists(fileName):
                return True
            if iTry >= maxTries:
                return False
            elif iTry == 1:
                logger.warning(
                    (
                        "Path `{}` does not exist (yet). We will try #{} more times with {}s "
                        "intervals."
                    ).format(fileName, maxTries - iTry, sleepTime)
                )
            time.sleep(sleepTime)
        return False

    def job_exists(self, job_id):
        return os.path.exists(self._get_job_file_name(job_id))

    def get_public_url(self, jobStoreFileID):
        self._check_job_store_file_id(jobStoreFileID)
        jobStorePath = self._get_file_path_from_id(jobStoreFileID)
        if os.path.exists(jobStorePath):
            return "file:" + jobStorePath
        else:
            raise NoSuchFileException(jobStoreFileID)

    def get_shared_public_url(self, sharedFileName):
        jobStorePath = os.path.join(self.sharedFilesDir, sharedFileName)
        if not os.path.exists(jobStorePath):
            raise NoSuchFileException(sharedFileName)
        return "file:" + jobStorePath

    def load_job(self, job_id):
        # If the job obviously doesn't exist, note that.
        self._check_job_store_id_exists(job_id)
        # Try to load a valid version of the job.
        jobFile = self._get_job_file_name(job_id)
        try:
            with open(jobFile, "rb") as fileHandle:
                job = pickle.load(fileHandle)
        except FileNotFoundError:
            # We were racing a delete on a non-POSIX-compliant filesystem.
            # This is the good case; the delete arrived in time.
            # If it didn't, we might go on to re-execute the already-finished job.
            # Anyway, this job doesn't really exist after all.
            raise NoSuchJobException(job_id)

        # Pass along the current config, which is the JobStore's responsibility.
        job.assignConfig(self.config)

        # The following cleans up any issues resulting from the failure of the
        # job during writing by the batch system.
        if os.path.isfile(jobFile + ".new"):
            logger.warning("There was a .new file for the job: %s", job_id)
            os.remove(jobFile + ".new")
            job.setupJobAfterFailure()
        return job

    def update_job(self, job):
        assert job.jobStoreID is not None, f"Tried to update job {job} without an ID"
        assert not isinstance(
            job.jobStoreID, TemporaryID
        ), f"Tried to update job {job} without an assigned ID"

        job.pre_update_hook()

        dest_filename = self._get_job_file_name(job.jobStoreID)

        # The job is serialised to a file suffixed by ".new"
        # We insist on creating the file; an existing .new file indicates
        # multiple simultaneous attempts to update the job, which will lose
        # updates.
        # The file is then moved to its correct path.
        # Atomicity guarantees use the fact the underlying file system's "move"
        # function is atomic.
        with open(dest_filename + ".new", "xb") as f:
            pickle.dump(job, f)
        # This should be atomic for the file system
        os.rename(dest_filename + ".new", dest_filename)

    def delete_job(self, job_id):
        # The jobStoreID is the relative path to the directory containing the job,
        # removing this directory deletes the job.
        if self.job_exists(job_id):
            # Remove the job-associated files in need of cleanup, which may or
            # may not live under the job's directory.
            robust_rmtree(self._get_job_files_cleanup_dir(job_id))
            # Remove the job's directory itself.
            robust_rmtree(self._get_job_dir_from_id(job_id))

    def jobs(self):
        # Walk through list of temporary directories searching for jobs.
        # Jobs are files that start with 'job'.
        # Note that this also catches jobWhatever.new which exists if an update
        # is in progress.
        for tempDir in self._job_directories():
            for i in os.listdir(tempDir):
                if i.startswith(self.JOB_DIR_PREFIX):
                    # This is a job instance directory
                    jobId = self._get_job_id_from_dir(os.path.join(tempDir, i))
                    try:
                        if self.job_exists(jobId):
                            yield self.load_job(jobId)
                    except NoSuchJobException:
                        # An orphaned job may leave an empty or incomplete job file which we can safely ignore
                        pass

    ##########################################
    # Functions that deal with temporary files associated with jobs
    ##########################################

    def _copy_or_link(self, src_path, dst_path, hardlink=False, symlink=False):
        # linking is not done be default because of issue #1755
        # TODO: is hardlinking ever actually done?
        src_path = self._extract_path_from_url(src_path)
        if self.linkImports and not hardlink and symlink:
            os.symlink(os.path.realpath(src_path), dst_path)
        else:
            atomic_copy(src_path, dst_path)

    def _import_file(
        self, otherCls, uri, shared_file_name=None, hardlink=False, symlink=True
    ):
        # symlink argument says whether the caller can take symlinks or not.
        # ex: if false, it means the workflow cannot work with symlinks and we need to hardlink or copy.
        # TODO: Do we ever actually hardlink?
        # default is true since symlinking everything is ideal
        uri_path = unquote(uri.path)
        if issubclass(otherCls, FileJobStore):
            if os.path.isdir(uri_path):
                # Don't allow directories (unless someone is racing us)
                raise IsADirectoryError(
                    f"URI {uri} points to a directory but a file was expected"
                )
            if shared_file_name is None:
                executable = os.stat(uri_path).st_mode & stat.S_IXUSR != 0
                # use this to get a valid path to write to in job store
                absPath = self._get_unique_file_path(uri_path)
                self._copy_or_link(uri, absPath, hardlink=hardlink, symlink=symlink)
                # TODO: os.stat(absPath).st_size consistently gives values lower than
                # getDirSizeRecursively()
                return FileID(
                    self._get_file_id_from_path(absPath),
                    os.stat(absPath).st_size,
                    executable,
                )
            else:
                self._requireValidSharedFileName(shared_file_name)
                path = self._get_shared_file_path(shared_file_name)
                self._copy_or_link(uri, path, hardlink=hardlink, symlink=symlink)
                return None
        else:
            return super()._import_file(
                otherCls, uri, shared_file_name=shared_file_name
            )

    def _export_file(self, otherCls, file_id, uri):
        if issubclass(otherCls, FileJobStore):
            srcPath = self._get_file_path_from_id(file_id)
            destPath = self._extract_path_from_url(uri)
            # Make sure we don't need to worry about directories when exporting
            # to local files, just like for cloud storage.
            os.makedirs(os.path.dirname(destPath), exist_ok=True)
            executable = getattr(file_id, "executable", False)
            if self.moveExports:
                self._move_and_linkback(srcPath, destPath, executable=executable)
            else:
                atomic_copy(srcPath, destPath, executable=executable)
        else:
            super()._default_export_file(otherCls, file_id, uri)

    def _move_and_linkback(self, srcPath, destPath, executable):
        logger.debug(
            "moveExports option, Moving src=%s to dest=%s ; then symlinking dest to src",
            srcPath,
            destPath,
        )
        shutil.move(srcPath, destPath)
        os.symlink(destPath, srcPath)
        if executable:
            os.chmod(destPath, os.stat(destPath).st_mode | stat.S_IXUSR)

    @classmethod
    def _url_exists(cls, url: ParseResult) -> bool:
        # Note that broken symlinks will not be shown to exist.
        return os.path.exists(cls._extract_path_from_url(url))

    @classmethod
    def _get_size(cls, url):
        return os.stat(cls._extract_path_from_url(url)).st_size

    @classmethod
    def _read_from_url(cls, url, writable):
        """
        Writes the contents of a file to a source (writes url to writable)
        using a ~10Mb buffer.

        :param str url: A path as a string of the file to be read from.
        :param object writable: An open file object to write to.
        """

        # we use a ~10Mb buffer to improve speed
        with cls._open_url(url) as readable:
            shutil.copyfileobj(readable, writable, length=cls.BUFFER_SIZE)
            # Return the number of bytes we read when we reached EOF.
            executable = os.stat(readable.name).st_mode & stat.S_IXUSR
            return readable.tell(), executable

    @classmethod
    def _open_url(cls, url: ParseResult) -> IO[bytes]:
        """
        Open a file URL as a binary stream.
        """
        return open(cls._extract_path_from_url(url), "rb")

    @classmethod
    def _write_to_url(cls, readable, url, executable=False):
        """
        Writes the contents of a file to a source (writes readable to url)
        using a ~10Mb buffer.

        :param str url: A path as a string of the file to be written to.
        :param object readable: An open file object to read from.
        """
        # we use a ~10Mb buffer to improve speed
        atomic_copyobj(
            readable,
            cls._extract_path_from_url(url),
            length=cls.BUFFER_SIZE,
            executable=executable,
        )

    @classmethod
    def _list_url(cls, url: ParseResult) -> list[str]:
        path = cls._extract_path_from_url(url)
        listing = []
        for p in os.listdir(path):
            # We know there are no slashes in these
            component = quote(p)
            # Return directories with trailing slashes and files without
            listing.append(
                (component + "/") if os.path.isdir(os.path.join(path, p)) else component
            )
        return listing

    @classmethod
    def _get_is_directory(cls, url: ParseResult) -> bool:
        path = cls._extract_path_from_url(url)
        return os.path.isdir(path)

    @staticmethod
    def _extract_path_from_url(url):
        """
        :return: local file path of file pointed at by the given URL
        """
        if url.netloc != "" and url.netloc != "localhost":
            raise RuntimeError("The URL '%s' is invalid" % url.geturl())
        return unquote(url.path)

    @classmethod
    def _supports_url(cls, url, export=False):
        return url.scheme.lower() == "file"

    def _make_string_filename_safe(self, arbitraryString, maxLength=240):
        """
        Given an arbitrary string, produce a filename-safe though not
        necessarily unique string based on it.

        The input string may be discarded altogether and replaced with any
        other nonempty filename-safe string.

        :param str arbitraryString: An arbitrary string
        :param int maxLength: Maximum length of the result, to keep it plus
                              any prefix or suffix under the filesystem's
                              path component length limit

        :return: A filename-safe string
        """

        # We will fill this in with the filename-safe parts we find.
        parts = []

        for substring in re.findall("[A-Za-z0-9._-]+", arbitraryString):
            # Collect all the matching substrings
            parts.append(substring)

        if len(parts) == 0:
            parts.append("UNPRINTABLE")

        # Glue it all together, and truncate to length
        return "_".join(parts)[:maxLength]

    def write_file(self, local_path, job_id=None, cleanup=False):
        absPath = self._get_unique_file_path(local_path, job_id, cleanup)
        relPath = self._get_file_id_from_path(absPath)
        atomic_copy(local_path, absPath)
        return relPath

    @contextmanager
    def write_file_stream(
        self, job_id=None, cleanup=False, basename=None, encoding=None, errors=None
    ):
        if not basename:
            basename = "stream"
        absPath = self._get_unique_file_path(basename, job_id, cleanup)
        relPath = self._get_file_id_from_path(absPath)

        with open(
            absPath,
            "wb" if encoding == None else "wt",
            encoding=encoding,
            errors=errors,
        ) as f:
            # Don't yield while holding an open file descriptor to the temp
            # file. That can result in temp files still being open when we try
            # to clean ourselves up, somehow, for certain workloads.
            yield f, relPath

    def get_empty_file_store_id(self, jobStoreID=None, cleanup=False, basename=None):
        with self.write_file_stream(jobStoreID, cleanup, basename) as (
            fileHandle,
            jobStoreFileID,
        ):
            return jobStoreFileID

    def update_file(self, file_id, local_path):
        self._check_job_store_file_id(file_id)
        jobStoreFilePath = self._get_file_path_from_id(file_id)

        if os.path.samefile(jobStoreFilePath, local_path):
            # The files are already the same file. We can't copy on eover the other.
            return

        atomic_copy(local_path, jobStoreFilePath)

    def read_file(self, file_id: str, local_path: str, symlink: bool = False) -> None:
        self._check_job_store_file_id(file_id)
        jobStoreFilePath = self._get_file_path_from_id(file_id)
        localDirPath = os.path.dirname(local_path)
        executable = getattr(file_id, "executable", False)

        if not symlink and os.path.islink(local_path):
            # We had a symlink and want to clobber it with a hardlink or copy.
            os.unlink(local_path)

        if os.path.exists(local_path) and os.path.samefile(
            jobStoreFilePath, local_path
        ):
            # The files are already the same: same name, hardlinked, or
            # symlinked. There is nothing to do, and trying to shutil.copyfile
            # one over the other will fail.
            return

        if symlink and self.symlink_job_store_reads:
            # If the reader will accept a symlink, and we are willing to
            # symlink into the jobstore, always give them one.
            # There's less that can go wrong.
            try:
                os.symlink(jobStoreFilePath, local_path)
                # It worked!
                return
            except OSError as e:
                # For the list of the possible errno codes, see: https://linux.die.net/man/2/symlink
                if e.errno == errno.EEXIST:
                    # Overwrite existing file, emulating shutil.copyfile().
                    os.unlink(local_path)
                    # It would be very unlikely to fail again for same reason but possible
                    # nonetheless in which case we should just give up.
                    os.symlink(jobStoreFilePath, local_path)
                    # Now we succeeded and don't need to copy
                    return
                elif e.errno == errno.EPERM:
                    # On some filesystems, the creation of symbolic links is not possible.
                    # In this case, we try to make a hard link.
                    pass
                else:
                    logger.error(
                        f"Unexpected OSError when reading file '{jobStoreFilePath}' from job store"
                    )
                    raise

        # If we get here, symlinking isn't an option.
        # Make sure we are working with the real source path, in case it is a
        # symlinked import.
        jobStoreFilePath = os.path.realpath(jobStoreFilePath)

        if os.stat(jobStoreFilePath).st_dev == os.stat(localDirPath).st_dev:
            # It is possible that we can hard link the file.
            # Note that even if the device numbers match, we can end up trying
            # to create a "cross-device" link.

            try:
                os.link(jobStoreFilePath, local_path)
                # It worked!
                return
            except OSError as e:
                # For the list of the possible errno codes, see: https://linux.die.net/man/2/link
                if e.errno == errno.EEXIST:
                    # Overwrite existing file, emulating shutil.copyfile().
                    os.unlink(local_path)
                    # It would be very unlikely to fail again for same reason but possible
                    # nonetheless in which case we should just give up.
                    os.link(jobStoreFilePath, local_path)
                    # Now we succeeded and don't need to copy
                    return
                elif e.errno == errno.EXDEV:
                    # It's a cross-device link even though it didn't appear to be.
                    # Just keep going and hit the file copy case.
                    pass
                elif e.errno == errno.EPERM:
                    # On some filesystems, hardlinking could be disallowed by permissions.
                    # In this case, we also fall back to making a complete copy.
                    pass
                elif e.errno == errno.ELOOP:
                    # Too many symbolic links were encountered. Just keep going and hit the
                    # file copy case.
                    pass
                elif e.errno == errno.EMLINK:
                    # The maximum number of links to file is reached. Just keep going and
                    # hit the file copy case.
                    pass
                else:
                    logger.error(
                        f"Unexpected OSError when reading file '{jobStoreFilePath}' from job store"
                    )
                    raise

        # If we get here, neither a symlink nor a hardlink will work.
        # Make a complete copy.
        atomic_copy(jobStoreFilePath, local_path, executable=executable)

    def delete_file(self, file_id):
        if not self.file_exists(file_id):
            return
        os.remove(self._get_file_path_from_id(file_id))

    def file_exists(self, file_id):
        absPath = self._get_file_path_from_id(file_id)

        if (
            not absPath.startswith(self.jobsDir)
            and not absPath.startswith(self.filesDir)
            and not absPath.startswith(self.jobFilesDir)
        ):
            # Don't even look for it, it is out of bounds.
            raise NoSuchFileException(file_id)

        try:
            st = os.stat(absPath)
        except OSError:
            return False
        if not stat.S_ISREG(st.st_mode):
            raise NoSuchFileException(file_id)
        return True

    def get_file_size(self, file_id):
        # Duplicate a bunch of fileExists to save on stat calls
        absPath = self._get_file_path_from_id(file_id)

        if (
            not absPath.startswith(self.jobsDir)
            and not absPath.startswith(self.filesDir)
            and not absPath.startswith(self.jobFilesDir)
        ):
            # Don't even look for it, it is out of bounds.
            raise NoSuchFileException(file_id)

        try:
            st = os.stat(absPath)
        except OSError:
            return 0
        return st.st_size

    @contextmanager
    def update_file_stream(self, file_id, encoding=None, errors=None):
        self._check_job_store_file_id(file_id)
        # File objects are context managers (CM) so we could simply return what open returns.
        # However, it is better to wrap it in another CM so as to prevent users from accessing
        # the file object directly, without a with statement.
        with open(
            self._get_file_path_from_id(file_id),
            "wb" if encoding == None else "wt",
            encoding=encoding,
            errors=errors,
        ) as f:
            yield f

    @contextmanager
    @overload
    def read_file_stream(
        self,
        file_id: Union[str, FileID],
        encoding: Literal[None] = None,
        errors: Optional[str] = None,
    ) -> Iterator[IO[bytes]]: ...

    @contextmanager
    @overload
    def read_file_stream(
        self, file_id: Union[str, FileID], encoding: str, errors: Optional[str] = None
    ) -> Iterator[IO[str]]: ...

    @contextmanager
    @overload
    def read_file_stream(
        self,
        file_id: Union[str, FileID],
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
    ) -> Union[Iterator[IO[bytes]], Iterator[IO[str]]]: ...

    @contextmanager
    def read_file_stream(
        self,
        file_id: Union[str, FileID],
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
    ) -> Union[Iterator[IO[bytes]], Iterator[IO[str]]]:
        self._check_job_store_file_id(file_id)
        if encoding is None:
            with open(
                self._get_file_path_from_id(file_id),
                "rb",
                encoding=encoding,
                errors=errors,
            ) as fb:
                yield fb
        else:
            with open(
                self._get_file_path_from_id(file_id),
                buffering=1,  # line buffering
                encoding=encoding,
                errors=errors,
            ) as ft:
                yield ft

    ##########################################
    # The following methods deal with shared files, i.e. files not associated
    # with specific jobs.
    ##########################################

    def _get_shared_file_path(self, sharedFileName):
        return os.path.join(self.sharedFilesDir, sharedFileName)

    @contextmanager
    def write_shared_file_stream(
        self, shared_file_name, encrypted=None, encoding=None, errors=None
    ):
        # the isProtected parameter has no effect on the fileStore
        self._requireValidSharedFileName(shared_file_name)
        with AtomicFileCreate(
            self._get_shared_file_path(shared_file_name)
        ) as tmpSharedFilePath:
            with open(
                tmpSharedFilePath,
                "wb" if encoding == None else "wt",
                encoding=encoding,
                errors=None,
            ) as f:
                yield f

    @overload
    @contextmanager
    def read_shared_file_stream(
        self,
        shared_file_name: str,
        encoding: str,
        errors: Optional[str] = None,
    ) -> Iterator[IO[str]]: ...

    @overload
    @contextmanager
    def read_shared_file_stream(
        self,
        shared_file_name: str,
        encoding: Literal[None] = None,
        errors: Optional[str] = None,
    ) -> Iterator[IO[bytes]]: ...

    @contextmanager
    def read_shared_file_stream(
        self,
        shared_file_name: str,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
    ) -> Union[Iterator[IO[bytes]], Iterator[IO[str]]]:
        self._requireValidSharedFileName(shared_file_name)
        try:
            with open(
                self._get_shared_file_path(shared_file_name),
                "rb" if encoding == None else "rt",
                encoding=encoding,
                errors=errors,
            ) as f:
                yield f

        except OSError as e:
            if e.errno == errno.ENOENT:
                raise NoSuchFileException(shared_file_name)
            else:
                raise

    def list_all_file_names(self, for_job: Optional[str] = None) -> Iterable[str]:
        """
        Get all the file names (not file IDs) of files stored in the job store.

        Used for debugging.

        :param for_job: If set, restrict the list to files for a particular job.
        """

        # TODO: Promote to AbstractJobStore.
        # TODO: Include stats-and-logging files?

        if for_job is not None:
            # Run on one job
            jobs = [for_job]
        else:
            # Run on all the jobs
            jobs = []
            # But not all the jobs that exist, we want all the jobs that have
            # files. So look at the file directories which mirror the job
            # directories' structure.
            for job_kind_dir in self._list_dynamic_spray_dir(self.jobFilesDir):
                # First we sprayed all the job kinds over a tree
                for job_instance_dir in self._list_dynamic_spray_dir(job_kind_dir):
                    # Then we sprayed the job instances over a tree
                    # And based on those we get the job name
                    job_id = self._get_job_id_from_files_dir(job_instance_dir)
                    jobs.append(job_id)

            yield from os.listdir(self.sharedFilesDir)

            for file_dir_path in self._list_dynamic_spray_dir(self.filesDir):
                # Run on all the no-job files
                yield from os.listdir(file_dir_path)

        for job_store_id in jobs:
            # Files from _get_job_files_dir
            job_files_dir = os.path.join(self.jobFilesDir, job_store_id)
            if os.path.exists(job_files_dir):
                for file_dir in os.listdir(job_files_dir):
                    # Each file is in its own directory
                    if file_dir == "cleanup":
                        # Except the cleanup directory which we do later.
                        continue
                    file_dir_path = os.path.join(job_files_dir, file_dir)
                    yield from os.listdir(file_dir_path)

                # Files from _get_job_files_cleanup_dir
                job_cleanup_files_dir = os.path.join(job_files_dir, "cleanup")
                if os.path.exists(job_cleanup_files_dir):
                    for file_dir in os.listdir(job_cleanup_files_dir):
                        # Each file is in its own directory
                        file_dir_path = os.path.join(job_cleanup_files_dir, file_dir)
                        yield from os.listdir(file_dir_path)

    def write_logs(self, msg):
        # Temporary files are placed in the stats directory tree
        tempStatsFileName = self.LOG_PREFIX + str(uuid.uuid4().hex) + self.LOG_TEMP_SUFFIX
        tempStatsFile = os.path.join(self._get_arbitrary_stats_inbox_dir(), tempStatsFileName)
        writeFormat = "w" if isinstance(msg, str) else "wb"
        with open(tempStatsFile, writeFormat) as f:
            f.write(msg)
        os.rename(tempStatsFile, tempStatsFile[:-len(self.LOG_TEMP_SUFFIX)])  # This operation is atomic

    def read_logs(self, callback, read_all=False):
        files_processed = 0

        # Holds pairs of a function to call to get directories to look at, and
        # a flag for whether to archive the files found.
        queries = []
        if read_all:
            # If looking at all logs, check the archive
            queries.append((self._stats_archive_directories, False))
        # Always check the inbox and archive from it. But do it after checking
        # the archive to avoid duplicates in the same pass.
        queries.append((self._stats_inbox_directories, True))

        for to_call, should_archive in queries:
            for log_dir in to_call():
                for log_file in os.listdir(log_dir):
                    if not log_file.startswith(self.LOG_PREFIX):
                        # Skip anything not a log file (like the other spray
                        # directories)
                        continue
                    if log_file.endswith(self.LOG_TEMP_SUFFIX):
                        # Skip partially-written files, always.
                        continue

                    abs_log_file = os.path.join(log_dir, log_file)
                    if not os.path.isfile(abs_log_file):
                        # This can't be a log file.
                        continue
                    try:
                        opened_file = open(abs_log_file, "rb")
                    except FileNotFoundError:
                        # File disappeared before we could open it.
                        # Maybe someone else is reading logs?
                        continue
                    with opened_file as f:
                        callback(f)
                    files_processed += 1

                    if should_archive:
                        # We need to move the stats file to the archive.
                        # Since we have UUID stats file names we don't need
                        # to worry about collisions when it gets there.
                        new_dir = self._get_arbitrary_stats_archive_dir()
                        new_abs_log_file = os.path.join(new_dir, log_file)
                        try:
                            # Mark this item as read
                            os.rename(abs_log_file, new_abs_log_file)
                        except FileNotFoundError:
                            # File we wanted to archive disappeared.
                            # Maybe someone else is reading logs?
                            # TODO: Raise ConcurrentFileModificationException?
                            continue
        return files_processed

    ##########################################
    # Private methods
    ##########################################

    def _get_job_dir_from_id(self, jobStoreID):
        """

        Find the directory for a job, which holds its job file.

        :param str jobStoreID: ID of a job, which is a relative to self.jobsDir.
        :rtype : string, string is the absolute path to a job directory inside self.jobsDir.
        """
        return os.path.join(self.jobsDir, jobStoreID)

    def _get_job_id_from_dir(self, absPath):
        """
        :param str absPath: The absolute path to a job directory under self.jobsDir which represents a job.
        :rtype : string, string is the job ID, which is a path relative to self.jobsDir
        """
        return absPath[len(self.jobsDir) + 1 :]

    def _get_job_id_from_files_dir(self, absPath: str) -> str:
        """
        :param str absPath: The absolute path to a job directory under self.jobFilesDir which holds a job's files.
        :rtype : string, string is the job ID
        """
        return absPath[len(self.jobFilesDir) + 1 :]

    def _get_job_file_name(self, jobStoreID):
        """
        Return the path to the file containing the serialised JobDescription instance for the given
        job.

        :rtype: str
        """
        return os.path.join(self._get_job_dir_from_id(jobStoreID), "job")

    def _get_job_files_dir(self, jobStoreID):
        """
        Return the path to the directory that should hold files made by the
        given job that should survive its deletion.

        This directory will only be created if files are to be put in it.

        :rtype : string, string is the absolute path to the job's files
                 directory
        """

        return os.path.join(self.jobFilesDir, jobStoreID)

    def _get_job_files_cleanup_dir(self, jobStoreID):
        """
        Return the path to the directory that should hold files made by the
        given job that will be deleted when the job is deleted.

        This directory will only be created if files are to be put in it.

        It may or may not be a subdirectory of the job's own directory.

        :rtype : string, string is the absolute path to the job's cleanup
                 files directory
        """

        return os.path.join(self.jobFilesDir, jobStoreID, "cleanup")

    def _check_job_store_id_assigned(self, jobStoreID):
        """
        Do nothing if the given job store ID has been assigned by
        :meth:`assignID`, and the corresponding job has not yet been
        deleted, even if the JobDescription hasn't yet been saved for the first
        time.

        If the ID has not been assigned, raises a NoSuchJobException.
        """

        if not self._wait_for_file(self._get_job_dir_from_id(jobStoreID)):
            raise NoSuchJobException(jobStoreID)

    def _check_job_store_id_exists(self, jobStoreID):
        """
        Raises a NoSuchJobException if the job with ID jobStoreID does not exist.
        """
        if not self._wait_for_exists(jobStoreID, 30):
            raise NoSuchJobException(jobStoreID)

    def _get_file_path_from_id(self, jobStoreFileID):
        """
        :param str jobStoreFileID: The ID of a file

        :rtype : string, string is the absolute path that that file should
                 appear at on disk, under either self.jobsDir if it is to be
                 cleaned up with a job, or self.filesDir otherwise.
        """

        # We just make the file IDs paths under the job store overall.
        absPath = os.path.join(self.jobStoreDir, unquote(jobStoreFileID))

        # Don't validate here, we are called by the validation logic

        return absPath

    def _get_file_id_from_path(self, absPath):
        """
        :param str absPath: The absolute path of a file.

        :rtype : string, string is the file ID.
        """

        return quote(absPath[len(self.jobStoreDir) + 1 :])

    def _check_job_store_file_id(self, jobStoreFileID):
        """
        :raise NoSuchFileException: if the file with ID jobStoreFileID does
                                    not exist or is not a file
        """
        if not self.file_exists(jobStoreFileID):
            raise NoSuchFileException(jobStoreFileID)

    def _get_arbitrary_jobs_dir_for_name(self, jobNameSlug):
        """
        Gets a temporary directory in a multi-level hierarchy in self.jobsDir.
        The directory is not unique and may already have other jobs' directories in it.
        We organize them at the top level by job name, to be user-inspectable.

        We make sure to prepend a string so that job names can't collide with
        spray directory names.

        :param str jobNameSlug: A partial filename derived from the job name.
                                Used as the first level of the directory hierarchy.

        :rtype : string, path to temporary directory in which to place files/directories.


        """

        if len(os.listdir(self.jobsDir)) > self.fanOut:
            # Make sure that we don't over-fill the root with too many unique job names.
            # Go in a subdirectory tree, and then go by job name and make another tree.
            return self._get_dynamic_spray_dir(
                os.path.join(
                    self._get_dynamic_spray_dir(self.jobsDir),
                    self.JOB_NAME_DIR_PREFIX + jobNameSlug,
                )
            )
        else:
            # Just go in the root
            return self._get_dynamic_spray_dir(
                os.path.join(self.jobsDir, self.JOB_NAME_DIR_PREFIX + jobNameSlug)
            )

    def _get_arbitrary_stats_inbox_dir(self):
        """
        Gets a temporary directory in a multi-level hierarchy in
        self.stats_inbox, where stats files not yet seen by the leader live.
        The directory is not unique and may already have other stats files in it.

        :rtype : string, path to temporary directory in which to place files/directories.


        """

        return self._get_dynamic_spray_dir(self.stats_inbox)

    def _get_arbitrary_stats_archive_dir(self):
        """
        Gets a temporary directory in a multi-level hierarchy in
        self.stats_archive, where stats files already seen by the leader live.
        The directory is not unique and may already have other stats files in it.

        :rtype : string, path to temporary directory in which to place files/directories.


        """

        return self._get_dynamic_spray_dir(self.stats_archive)

    def _get_arbitrary_files_dir(self):
        """
        Gets a temporary directory in a multi-level hierarchy in self.filesDir.
        The directory is not unique and may already have other user files in it.

        :rtype : string, path to temporary directory in which to place files/directories.


        """

        return self._get_dynamic_spray_dir(self.filesDir)

    def _get_dynamic_spray_dir(self, root):
        """
        Gets a temporary directory in a possibly multi-level hierarchy of
        directories under the given root.

        Each time a directory in the hierarchy starts to fill up, additional
        hierarchy levels are created under it, and we randomly "spray" further
        files and directories across them.

        We can't actually enforce that we never go over our internal limit for
        files in a directory, because any number of calls to this function can
        be happening simultaneously. But we can enforce that, once too many
        files are visible on disk, only subdirectories will be created.

        The returned directory will exist, and may contain other data already.

        The caller may not create any files or directories in the returned
        directory with single-character names that are in self.validDirs.

        :param str root : directory to put the hierarchy under, which will
                          fill first.

        :rtype : string, path to temporary directory in which to place
                 files/directories.
        """
        tempDir = root

        # Make sure the root exists
        os.makedirs(tempDir, exist_ok=True)

        while len(os.listdir(tempDir)) >= self.fanOut:
            # We need to use a layer of directories under here to avoid over-packing the directory
            tempDir = os.path.join(tempDir, random.choice(self.validDirs))
            os.makedirs(tempDir, exist_ok=True)

        # When we get here, we found a sufficiently empty directory
        return tempDir

    def _walk_dynamic_spray_dir(self, root):
        """
        Walks over a directory tree filled in by _getDynamicSprayDir.

        Yields each directory _getDynamicSprayDir has ever returned, and no
        directories it has not returned (besides the root).

        If the caller looks in the directory, they must ignore subdirectories
        with single-character names in self.validDirs.

        :param str root : directory the hierarchy was put under

        :rtype : an iterator over directories
        """

        # Always yield the root.
        # The caller is responsible for dealing with it if it has gone away.
        yield root

        children = []

        try:
            # Look for children
            children = os.listdir(root)
        except:
            # Don't care if they are gone
            pass

        for child in children:
            # Go over all the children
            if child not in self.validDirsSet:
                # Only look at our reserved names we use for fan-out
                continue

            # We made this directory, so go look in it
            childPath = os.path.join(root, child)

            # Recurse
            yield from self._walk_dynamic_spray_dir(childPath)

    def _list_dynamic_spray_dir(self, root):
        """
        For a directory tree filled in by _getDynamicSprayDir, yields each
        highest-level file or or directory *not* created by _getDynamicSprayDir
        (i.e. the actual contents).
        """

        for spray_dir in self._walk_dynamic_spray_dir(root):
            for child in os.listdir(spray_dir):
                if child not in self.validDirsSet:
                    # This is a real content item we are storing
                    yield os.path.join(spray_dir, child)

    def _job_directories(self):
        """
        :rtype : an iterator to the temporary directories containing job
                 files. They may also contain directories containing more
                 job files.
        """

        # Walking the job directories is more complicated.
        # We have one layer of spray (which is sometimes bypassed, but that's OK), then a job name, then another layer.
        # We can tell the job name directories from the spray directories because they start with self.JOB_NAME_DIR_PREFIX.
        # We never look at the directories containing the job name directories,
        # so they aren't mistaken for the leaf-level per-job job directories.

        for jobHoldingDir in self._walk_dynamic_spray_dir(self.jobsDir):
            # For every directory in the first spray, look at children
            children = []

            try:
                children = os.listdir(jobHoldingDir)
            except:
                pass

            for jobNameDir in children:
                if not jobNameDir.startswith(self.JOB_NAME_DIR_PREFIX):
                    continue

                # Now we have only the directories that are named after jobs. Look inside them.
                yield from self._walk_dynamic_spray_dir(
                    os.path.join(jobHoldingDir, jobNameDir)
                )

    def _stats_inbox_directories(self):
        """
        :returns: an iterator to the temporary directories containing new stats
            files. They may also contain directories containing more stats
            files.
        """

        return self._walk_dynamic_spray_dir(self.stats_inbox)

    def _stats_archive_directories(self):
        """
        :returns: an iterator to the temporary directories containing
            previously observed stats files. They may also contain directories
            containing more stats files.
        """

        return self._walk_dynamic_spray_dir(self.stats_archive)

    def _get_unique_file_path(self, fileName, jobStoreID=None, cleanup=False):
        """
        Create unique file name within a jobStore directory or tmp directory.

        :param fileName: A file name, which can be a full path as only the
        basename will be used.
        :param jobStoreID: If given, the path returned will be in a directory including the job's ID as part of its path.
        :param bool cleanup: If True and jobStoreID is set, the path will be in
            a place such that it gets deleted when the job is deleted.
        :return: The full path with a unique file name.
        """

        # Give the file a unique directory that either will be cleaned up with a job or won't.
        directory = self._get_file_directory(jobStoreID, cleanup)
        # And then a path under it
        uniquePath = os.path.join(directory, os.path.basename(fileName))
        # No need to check if it exists already; it is in a unique directory.
        return uniquePath

    def _get_file_directory(self, jobStoreID=None, cleanup=False):
        """
        Get a new empty directory path for a file to be stored at.


        :param str jobStoreID: If the jobStoreID is not None, the file wil
               be associated with the job with that ID.

        :param bool cleanup: If cleanup is also True, this directory
               will be cleaned up when the job is deleted.

        :rtype :string, string is the absolute path to a directory to put the file in.
        """
        if jobStoreID != None:
            # Make a temporary file within the job's files directory

            # Make sure the job is legit
            self._check_job_store_id_assigned(jobStoreID)
            # Find where all its created files should live, depending on if
            # they need to go away when the job is deleted or not.
            jobFilesDir = (
                self._get_job_files_dir(jobStoreID)
                if not cleanup
                else self._get_job_files_cleanup_dir(jobStoreID)
            )

            # Lazily create the parent directory.
            # We don't want our tree filled with confusingly empty directories.
            os.makedirs(jobFilesDir, exist_ok=True)

            # Then make a temp directory inside it
            filesDir = os.path.join(jobFilesDir, "file-" + uuid.uuid4().hex)
            os.mkdir(filesDir)
            return filesDir
        else:
            # Make a temporary file within the non-job-associated files hierarchy
            filesDir = os.path.join(
                self._get_arbitrary_files_dir(), "file-" + uuid.uuid4().hex
            )
            os.mkdir(filesDir)
            return filesDir



import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp


def binaryStrings(job, depth, message=""):
    if depth > 0:
        s = [
            job.addChildJobFn(binaryStrings, depth - 1, message + "0").rv(),
            job.addChildJobFn(binaryStrings, depth - 1, message + "1").rv(),
        ]
        return job.addFollowOnFn(merge, s).rv()
    return [message]


def merge(strings):
    return strings[0] + strings[1]


if __name__ == "__main__":
    jobstore: str = mkdtemp("tutorial_promises2")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.loglevel = "OFF"
    options.clean = "always"

    with Toil(options) as toil:
        print(toil.start(Job.wrapJobFn(binaryStrings, depth=5)))



from argparse import ArgumentParser

from toil.lib.conversions import human2bytes


def add_runner_options(
    parser: ArgumentParser, cwl: bool = False, wdl: bool = False
) -> None:
    """
    Add to the WDL or CWL runners options that are shared or the same between runners
    :param parser: parser to add arguments to
    :param cwl: bool
    :param wdl: bool
    :return: None
    """
    # This function should be constructed so that even when wdl and cwl are false, the "default" options are still added
    run_imports_on_workers_arguments = ["--runImportsOnWorkers"]
    if cwl:
        run_imports_on_workers_arguments.append("--run-imports-on-workers")
    parser.add_argument(
        *run_imports_on_workers_arguments,
        action="store_true",
        default=False,
        dest="run_imports_on_workers",
        help="Run the file imports on a worker instead of the leader. This is useful if the leader is not optimized for high network performance. "
        "If set to true, the argument --importWorkersDisk must also be set."
    )
    import_workers_threshold_argument = ["--importWorkersThreshold"]
    if cwl:
        import_workers_threshold_argument.append("--import-workers-threshold")
    parser.add_argument(
        *import_workers_threshold_argument,
        dest="import_workers_threshold",
        type=lambda x: human2bytes(str(x)),
        default="1 GiB",
        help="Specify the file size threshold that determines how many files go into a batched import. As many files will go into a batch import job until this threshold "
             "is reached. This should be set in conjunction with the argument --runImportsOnWorkers."
    )
    import_workers_disk_argument = ["--importWorkersDisk"]
    if cwl:
        import_workers_disk_argument.append("--import-workers-disk")
    parser.add_argument(
        *import_workers_disk_argument,
        dest="import_workers_disk",
        type=lambda x: human2bytes(str(x)),
        default="1 MiB",
        help="Specify the disk size each import worker will get. This may be necessary when file streaming is not possible. For example, downloading from AWS to a GCE "
             "job store. If specified, this should be set to the largest file size of all files to import. This should be set in conjunction with the arguments "
             "--runImportsOnWorkers and --importWorkersThreshold."
    )



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from docker.errors import ImageNotFound
from toil import checkDockerImageExists, parseDockerAppliance
from toil.lib.retry import retry
from toil.test import ToilTest, needs_docker


@needs_docker
class DockerCheckTest(ToilTest):
    """Tests checking whether a docker image exists or not."""

    @unittest.skip("Consumes unauthenticated Docker Hub pulls if run")
    def testOfficialUbuntuRepo(self):
        """Image exists.  This should pass."""
        ubuntu_repo = "ubuntu:latest"
        assert checkDockerImageExists(ubuntu_repo)

    @unittest.skip("Consumes unauthenticated Docker Hub pulls if run")
    def testBroadDockerRepo(self):
        """Image exists.  This should pass."""
        broad_repo = "broadinstitute/genomes-in-the-cloud:2.0.0"
        assert checkDockerImageExists(broad_repo)

    @unittest.skip("Consumes unauthenticated Docker Hub pulls if run")
    def testBroadDockerRepoBadTag(self):
        """Bad tag.  This should raise."""
        broad_repo = "broadinstitute/genomes-in-the-cloud:-----"
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(broad_repo)

    @unittest.skip("Consumes unauthenticated Docker Hub pulls if run")
    def testNonexistentRepo(self):
        """Bad image.  This should raise."""
        nonexistent_repo = "------:-----"
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_repo)

    def testToilQuayRepo(self):
        """Image exists.  Should pass."""
        toil_repo = "quay.io/ucsc_cgl/toil:latest"
        assert checkDockerImageExists(toil_repo)

    def testBadQuayRepoNTag(self):
        """Bad repo and tag.  This should raise."""
        nonexistent_quay_repo = "quay.io/--------:---"
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_quay_repo)

    def testBadQuayRepo(self):
        """Bad repo.  This should raise."""
        nonexistent_quay_repo = "quay.io/--------:latest"
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_quay_repo)

    def testBadQuayTag(self):
        """Bad tag.  This should raise."""
        nonexistent_quay_repo = "quay.io/ucsc_cgl/toil:---"
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_quay_repo)

    def testGoogleRepo(self):
        """Image exists.  Should pass."""
        google_repo = "gcr.io/google-containers/busybox:latest"
        assert checkDockerImageExists(google_repo)

    @retry(
        errors=[TimeoutError]
    )  # see: https://github.com/DataBiosphere/toil/issues/4902
    def testBadGoogleRepo(self):
        """Bad repo and tag.  This should raise."""
        nonexistent_google_repo = "gcr.io/google-containers/--------:---"
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_google_repo)

    def testApplianceParser(self):
        """Test that a specified appliance is parsed correctly."""
        docker_list = [
            "ubuntu:latest",
            "ubuntu",
            "broadinstitute/genomes-in-the-cloud:2.0.0",
            "quay.io/ucsc_cgl/toil:latest",
            "gcr.io/google-containers/busybox:latest",
        ]
        parsings = []
        for image in docker_list:
            registryName, imageName, tag = parseDockerAppliance(image)
            parsings.append([registryName, imageName, tag])
        expected_parsings = [
            ["docker.io", "ubuntu", "latest"],
            ["docker.io", "ubuntu", "latest"],
            ["docker.io", "broadinstitute/genomes-in-the-cloud", "2.0.0"],
            ["quay.io", "ucsc_cgl/toil", "latest"],
            ["gcr.io", "google-containers/busybox", "latest"],
        ]
        assert parsings == expected_parsings



from typing import List

from _typeshed import Incomplete

from ..core.listener import IListenerExcHandler
from ..core.topicmgr import TopicManager

class TracebackInfo:
    ExcClass: Incomplete
    excArg: Incomplete
    traceback: Incomplete
    def __init__(self) -> None: ...
    def getFormattedList(self) -> List[str]: ...
    def getFormattedString(self) -> str: ...

class ExcPublisher(IListenerExcHandler):
    topicUncaughtExc: str
    def __init__(self, topicMgr: TopicManager = ...) -> None: ...
    def init(self, topicMgr: TopicManager): ...
    def __call__(self, listenerID: str, topicObj): ...



from typing import Any, Iterator, List, Sequence, TextIO, Tuple, Union

from _typeshed import Incomplete

from .topicargspec import ArgsDocs, ArgSpecGiven, topicArgsFromCallable
from .topicexc import UnrecognizedSourceFormatError as UnrecognizedSourceFormatError
from .topicobj import Topic
from .topictreetraverser import TopicTreeTraverser

class ITopicDefnProvider:
    def getDefn(self, topicNameTuple: Sequence[str]) -> Tuple[str, ArgSpecGiven]: ...
    def topicNames(self) -> Sequence[str]: ...
    def getTreeDoc(self) -> str: ...
    def __iter__(self) -> Iterator[str]: ...

SPEC_METHOD_NAME: str

class ITopicDefnDeserializer:
    class TopicDefn:
        nameTuple: Incomplete
        description: Incomplete
        argsDocs: Incomplete
        required: Incomplete
        def __init__(
            self,
            nameTuple: Sequence[str],
            description: str,
            argsDocs: ArgsDocs,
            required: Sequence[str],
        ) -> None: ...
        def isComplete(self) -> bool: ...

    def getTreeDoc(self) -> str: ...
    def getNextTopic(self) -> TopicDefn: ...
    def doneIter(self) -> None: ...
    def resetIter(self) -> None: ...

class TopicDefnDeserialClass(ITopicDefnDeserializer):
    def __init__(self, pyClassObj: type = ...) -> None: ...
    def getTreeDoc(self) -> str: ...
    def getNextTopic(self) -> ITopicDefnDeserializer.TopicDefn: ...
    def resetIter(self) -> None: ...
    def getDefinedTopics(self) -> List[str]: ...

class TopicDefnDeserialModule(ITopicDefnDeserializer):
    def __init__(self, moduleName: str, searchPath: Sequence[str] = ...) -> None: ...
    def getTreeDoc(self) -> str: ...
    def getNextTopic(self) -> ITopicDefnDeserializer.TopicDefn: ...
    def doneIter(self) -> None: ...
    def resetIter(self) -> None: ...
    def getDefinedTopics(self) -> List[str]: ...

class TopicDefnDeserialString(ITopicDefnDeserializer):
    def __init__(self, source: str) -> None: ...
    def getTreeDoc(self) -> str: ...
    def getNextTopic(self) -> ITopicDefnDeserializer.TopicDefn: ...
    def doneIter(self) -> None: ...
    def resetIter(self) -> None: ...
    def getDefinedTopics(self) -> List[str]: ...

TOPIC_TREE_FROM_MODULE: str
TOPIC_TREE_FROM_STRING: str
TOPIC_TREE_FROM_CLASS: str

class TopicDefnProvider(ITopicDefnProvider):
    def __init__(self, source: Any, format: str, **providerKwargs: Any) -> None: ...
    def getDefn(self, topicNameTuple: Sequence[str]) -> Tuple[str, ArgSpecGiven]: ...
    def topicNames(self) -> Sequence[str]: ...
    def getTreeDoc(self) -> str: ...
    @classmethod
    def registerTypeForImport(cls, typeName: str, providerClassObj: type) -> None: ...
    @classmethod
    def initTypeRegistry(cls) -> None: ...

defaultTopicTreeSpecHeader: str
defaultTopicTreeSpecFooter: str

def exportTopicTreeSpec(
    moduleName: str = ...,
    rootTopic: Union[Topic, str] = ...,
    bak: str = ...,
    moduleDoc: str = ...,
) -> None: ...

class TopicTreeSpecPrinter:
    INDENT_CH: str
    def __init__(
        self,
        rootTopic: Union[str, Topic] = ...,
        fileObj: TextIO = ...,
        width: int = ...,
        indentStep: int = ...,
        treeDoc: str = ...,
        footer: str = ...,
    ): ...
    def getOutput(self) -> str: ...
    def writeAll(self, topicObj: Topic) -> None: ...



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from toil.job import Job
from toil.test import ToilTest


class CachedUnpicklingJobStoreTest(ToilTest):
    # https://github.com/BD2KGenomics/toil/issues/817
    def test(self):
        """
        Runs two identical Toil workflows with different job store paths
        """
        for _ in range(2):
            options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
            options.logLevel = "INFO"
            root = Job.wrapJobFn(parent)
            Job.Runner.startToil(root, options)


def parent(job):
    return job.addChildFn(child).rv()


def child():
    return 1


class ChainedIndexedPromisesTest(ToilTest):
    # https://github.com/BD2KGenomics/toil/issues/1021
    def test(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.logLevel = "INFO"
        root = Job.wrapJobFn(a)
        self.assertEqual(Job.Runner.startToil(root, options), 42)


def a(job):
    return job.addChild(job.wrapJobFn(b)).rv(0)


def b(job):
    return job.addChild(job.wrapFn(c)).rv()


def c():
    return 42, 43


class PathIndexingPromiseTest(ToilTest):
    """
    Test support for indexing promises of arbitrarily nested data structures of lists, dicts and
    tuples, or any other object supporting the __getitem__() protocol.
    """

    def test(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.logLevel = "INFO"
        root = Job.wrapJobFn(d)
        self.assertEqual(Job.Runner.startToil(root, options), ("b", 43, 3))


def d(job):
    child = job.addChild(job.wrapFn(e))
    return child.rv("a"), child.rv(42), child.rv("c", 2)


def e():
    return {"a": "b", 42: 43, "c": [1, 2, 3]}



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import builtins
from collections.abc import Callable, Generator
import logging
import os
from pathlib import Path
import re
import subprocess
import sys
import time
import uuid
from typing import Optional, Any, cast
import pytest

import toil
from toil import resolveEntryPoint
from toil.common import Config, Toil
from toil.job import Job
from toil.lib.bioio import system
from toil.fileStores.abstractFileStore import AbstractFileStore
from toil.test import (
    get_data,
    pneeds_aws_ec2 as needs_aws_ec2,
    pneeds_cwl as needs_cwl,
    pneeds_docker as needs_docker,
    pintegrative as integrative,
    pneeds_rsync3 as needs_rsync3,
    pslow as slow,
)
from toil.test.sort.sort import makeFileToSort
from toil.utils.toilStats import get_stats, process_data
from toil.utils.toilStatus import ToilStatus
from toil.version import python

logger = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def unsortedFile(tmp_path: Path) -> Generator[Path]:
    try:
        tempFile = tmp_path / "lines"
        makeFileToSort(str(tempFile), 1000, 10)
        yield tempFile
    finally:
        pass  # no cleanup needed


@pytest.fixture(scope="function")
def correctSort(unsortedFile: Path) -> list[str]:
    with unsortedFile.open() as fileHandle:
        lines = fileHandle.readlines()
        lines.sort()
        return lines


class TestUtils:
    """
    Tests the utilities that toil ships with, e.g. stats and status, in conjunction with restart
    functionality.
    """

    N: int = 1000

    @property
    def toilMain(self) -> str:
        return resolveEntryPoint("toil")

    def cleanCommand(self, jobstore: Path) -> list[str]:
        return [self.toilMain, "clean", str(jobstore)]

    def statsCommand(self, jobstore: Path) -> list[str]:
        return [self.toilMain, "stats", str(jobstore), "--pretty"]

    def statusCommand(
        self, jobstore: Path, failIfNotComplete: bool = False
    ) -> list[str]:
        commandTokens = [self.toilMain, "status", str(jobstore)]
        if failIfNotComplete:
            commandTokens.append("--failIfNotComplete")
        return commandTokens

    def test_config_functionality(self, tmp_path: Path) -> None:
        """Ensure that creating and reading back the config file works"""
        config_file = tmp_path / "config.yaml"
        config_command = [self.toilMain, "config", str(config_file)]
        # make sure the command `toil config file_path` works
        try:
            subprocess.check_call(config_command)
        except subprocess.CalledProcessError:
            pytest.fail("The toil config utility failed!")

        parser = Job.Runner.getDefaultArgumentParser()
        # make sure that toil can read from the generated config file
        try:
            parser.parse_args(
                [str(tmp_path / "random_jobstore"), "--config", str(config_file)]
            )
            with config_file.open() as cm:
                payload = cm.read()
                expected = "workDir batchSystem symlinkImports defaultMemory retryCount"
                assert all(
                    re.search(rf"^#*{ param }:", payload, re.MULTILINE)
                    for param in expected.split(" ")
                ), f"Generated config contains { expected }"
        except SystemExit:
            pytest.fail("Failed to parse the default generated config file!")

    @needs_rsync3
    @pytest.mark.timeout(1200)
    @needs_aws_ec2
    @integrative
    @pytest.mark.rsync
    @pytest.mark.online
    @pytest.mark.aws_s3
    @pytest.mark.integrative
    @slow
    @pytest.mark.slow
    def testAWSProvisionerUtils(self) -> None:
        """
        Runs a number of the cluster utilities in sequence.

        Launches a cluster with custom tags.
        Verifies the tags exist.
        ssh's into the cluster.
        Does some weird string comparisons.
        Makes certain that TOIL_WORKDIR is set as expected in the ssh'ed cluster.
        Rsyncs a file and verifies it exists on the leader.
        Destroys the cluster.

        :return:
        """
        # TODO: Run these for the other clouds.
        clusterName = f"cluster-utils-test{uuid.uuid4()}"
        keyName = os.getenv("TOIL_AWS_KEYNAME", "id_rsa").strip()
        expected_owner = os.getenv("TOIL_OWNER_TAG", keyName)

        try:
            from toil.provisioners.aws.awsProvisioner import AWSProvisioner

            aws_provisioner = AWSProvisioner.__module__
            logger.debug(f"Found AWSProvisioner: {aws_provisioner}.")

            # launch master with an assortment of custom tags
            system(
                [
                    self.toilMain,
                    "launch-cluster",
                    "--clusterType",
                    "mesos",
                    "-t",
                    "key1=value1",
                    "-t",
                    "key2=value2",
                    "--tag",
                    "key3=value3",
                    "--leaderNodeType=t2.medium",
                    "--keyPairName=" + keyName,
                    clusterName,
                    "--provisioner=aws",
                    "--zone=us-west-2a",
                    "--logLevel=DEBUG",
                ]
            )

            cluster = toil.provisioners.cluster_factory(
                provisioner="aws", zone="us-west-2a", clusterName=clusterName
            )
            leader = cluster.getLeader()

            # check that the leader carries the appropriate tags
            tags = {
                "key1": "value1",
                "key2": "value2",
                "key3": "value3",
                "Name": clusterName,
                "Owner": expected_owner,
            }
            for key in tags:
                assert cast(dict[str, str], leader.tags).get(key) == tags[key]
        finally:
            system(
                [
                    self.toilMain,
                    "destroy-cluster",
                    "--zone=us-west-2a",
                    "--provisioner=aws",
                    clusterName,
                ]
            )

    @slow
    def testUtilsSort(
        self, tmp_path: Path, unsortedFile: Path, correctSort: list[str]
    ) -> None:
        """
        Tests the status and stats commands of the toil command line utility using the
        sort example with the --restart flag.
        """
        jobstore = tmp_path / "jobstore"
        outputFile = tmp_path / "someSortedStuff.txt"
        # Get the sort command to run
        toilCommand = [
            sys.executable,
            "-m",
            toil.test.sort.sort.__name__,
            str(jobstore),
            "--logLevel=DEBUG",
            "--fileToSort",
            str(unsortedFile),
            "--outputFile",
            str(outputFile),
            "--N",
            str(self.N),
            "--stats",
            "--retryCount=2",
            "--badWorker=0.5",
            "--badWorkerFailInterval=0.05",
        ]
        # Try restarting it to check that a JobStoreException is thrown
        with pytest.raises(subprocess.CalledProcessError):
            system(toilCommand + ["--restart"])
        # Check that trying to run it in restart mode does not create the jobStore
        assert not jobstore.exists()

        # Status command
        # Run the script for the first time
        try:
            system(toilCommand)
            finished = True
        except (
            subprocess.CalledProcessError
        ):  # This happens when the script fails due to having unfinished jobs
            system(self.statusCommand(jobstore))
            with pytest.raises(subprocess.CalledProcessError):
                system(self.statusCommand(jobstore, failIfNotComplete=True))
            finished = False
        assert jobstore.exists()

        # Try running it without restart and check an exception is thrown
        with pytest.raises(subprocess.CalledProcessError):
            system(toilCommand)

        # Now restart it until done
        totalTrys = 1
        while not finished:
            try:
                system(toilCommand + ["--restart"])
                finished = True
            except (
                subprocess.CalledProcessError
            ):  # This happens when the script fails due to having unfinished jobs
                system(self.statusCommand(jobstore))
                with pytest.raises(subprocess.CalledProcessError):
                    system(self.statusCommand(jobstore, failIfNotComplete=True))
                if totalTrys > 16:
                    pytest.fail("Exceeded a reasonable number of restarts")
                totalTrys += 1

        # Check the toil status command does not issue an exception
        system(self.statusCommand(jobstore))

        # Check we can run 'toil stats'
        system(self.statsCommand(jobstore))

        # Check the file is properly sorted
        with outputFile.open() as fileHandle:
            l2 = fileHandle.readlines()
            assert correctSort == l2

        # Check we can run 'toil clean'
        system(self.cleanCommand(jobstore))

    @slow
    @pytest.mark.slow
    def testUtilsStatsSort(
        self, tmp_path: Path, unsortedFile: Path, correctSort: list[str]
    ) -> None:
        """
        Tests the stats commands on a complete run of the stats test.
        """
        jobstore = tmp_path / "jobstore"
        outputFile = tmp_path / "someSortedStuff.txt"
        # Get the sort command to run
        toilCommand = [
            sys.executable,
            "-m",
            toil.test.sort.sort.__name__,
            str(jobstore),
            "--logLevel=DEBUG",
            "--fileToSort",
            str(unsortedFile),
            "--outputFile",
            str(outputFile),
            "--N",
            str(self.N),
            "--stats",
            "--retryCount=99",
            "--badWorker=0.5",
            "--badWorkerFailInterval=0.01",
        ]

        # Run the script for the first time
        system(toilCommand)
        assert jobstore.exists()

        # Check we can run 'toil stats'
        system(self.statsCommand(jobstore))

        # Check the file is properly sorted
        with outputFile.open() as fileHandle:
            l2 = fileHandle.readlines()
            assert correctSort == l2

    def testUnicodeSupport(self, tmp_path: Path) -> None:
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.clean = "always"
        options.logLevel = "debug"
        Job.Runner.startToil(Job.wrapFn(printUnicodeCharacter), options)

    @slow
    @pytest.mark.slow
    def testMultipleJobsPerWorkerStats(self, tmp_path: Path) -> None:
        """
        Tests case where multiple jobs are run on 1 worker to ensure that all jobs report back their data
        """
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.clean = "never"
        options.stats = True
        Job.Runner.startToil(RunTwoJobsPerWorker(), options)
        config = Config()
        config.setOptions(options)
        jobStore = Toil.resumeJobStore(config.jobStore)
        stats = get_stats(jobStore)
        collatedStats = process_data(jobStore.config, stats)
        assert (
            len(collatedStats.job_types) == 2  # type: ignore[attr-defined]
        ), "Some jobs are not represented in the stats."

    def check_status(
        self,
        jobstore: Path,
        status: str,
        status_fn: Callable[[str], str],
        process: Optional[subprocess.Popen[Any]] = None,
        seconds: int = 20,
    ) -> None:
        time_elapsed = 0.0
        has_stopped = process.poll() is not None if process else False
        current_status = status_fn(str(jobstore))
        while current_status != status:
            if has_stopped:
                # If the process has stopped and the stratus is wrong, it will never be right.
                assert process is not None
                assert (
                    current_status == status
                ), f"Process returned {process.returncode} without status reaching {status}; stuck at {current_status}"
            logger.debug(
                "Workflow is %s; waiting for %s (%s/%s elapsed)",
                current_status,
                status,
                time_elapsed,
                seconds,
            )
            time.sleep(0.5)
            time_elapsed += 0.5
            has_stopped = process.poll() is not None if process else False
            current_status = status_fn(str(jobstore))
            if time_elapsed > seconds:
                assert (
                    current_status == status
                ), "Waited {seconds} seconds without status reaching {status}; stuck at {current_status}"

    def testGetPIDStatus(self, tmp_path: Path, unsortedFile: Path) -> None:
        """Test that ToilStatus.getPIDStatus() behaves as expected."""
        jobstore = tmp_path / "jobstore"
        outputFile = tmp_path / "someSortedStuff.txt"
        wf = subprocess.Popen(
            [
                python,
                "-m",
                "toil.test.sort.sort",
                jobstore.as_uri(),
                f"--fileToSort={unsortedFile}",
                f"--outputFile={outputFile}",
                "--clean=never",
            ]
        )
        self.check_status(
            jobstore,
            "RUNNING",
            status_fn=ToilStatus.getPIDStatus,
            process=wf,
            seconds=60,
        )
        wf.wait()
        self.check_status(
            jobstore,
            "COMPLETED",
            status_fn=ToilStatus.getPIDStatus,
            process=wf,
            seconds=60,
        )

        # TODO: we need to reach into the FileJobStore's files and delete this
        #  shared file. We assume we know its internal layout.
        os.remove(jobstore / "files/shared/pid.log")
        self.check_status(
            jobstore,
            "QUEUED",
            status_fn=ToilStatus.getPIDStatus,
            process=wf,
            seconds=60,
        )

    def testGetStatusFailedToilWF(self, tmp_path: Path, unsortedFile: Path) -> None:
        """
        Test that ToilStatus.getStatus() behaves as expected with a failing Toil workflow.
        While this workflow could be called by importing and evoking its main function, doing so would remove the
        opportunity to test the 'RUNNING' functionality of getStatus().
        """
        # --badWorker is set to force failure.
        jobstore = tmp_path / "jobstore"
        outputFile = tmp_path / "someSortedStuff.txt"
        wf = subprocess.Popen(
            [
                python,
                "-m",
                "toil.test.sort.sort",
                jobstore.as_uri(),
                f"--fileToSort={unsortedFile}",
                f"--outputFile={outputFile}",
                "--clean=never",
                "--badWorker=1",
            ]
        )
        self.check_status(
            jobstore, "RUNNING", status_fn=ToilStatus.getStatus, process=wf, seconds=60
        )
        wf.wait()
        self.check_status(
            jobstore, "ERROR", status_fn=ToilStatus.getStatus, process=wf, seconds=60
        )

    @needs_cwl
    @needs_docker
    @pytest.mark.cwl
    @pytest.mark.docker
    @pytest.mark.online
    def testGetStatusFailedCWLWF(self, tmp_path: Path) -> None:
        """Test that ToilStatus.getStatus() behaves as expected with a failing CWL workflow."""
        jobstore = tmp_path / "jobstore"
        with get_data("test/cwl/sorttool.cwl") as cwl_file:
            with get_data("test/cwl/whale.txt") as input_file:
                # --badWorker is set to force failure.
                cmd = [
                    "toil-cwl-runner",
                    "--logDebug",
                    "--jobStore",
                    str(jobstore),
                    "--clean=never",
                    "--badWorker=1",
                    str(cwl_file),
                    "--reverse",
                    "--input",
                    str(input_file),
                    f"--outdir={tmp_path}",
                ]
                logger.info("Run command: %s", " ".join(cmd))
                wf = subprocess.Popen(cmd)
                self.check_status(
                    jobstore,
                    "RUNNING",
                    status_fn=ToilStatus.getStatus,
                    process=wf,
                    seconds=60,
                )
                wf.wait()
                self.check_status(
                    jobstore,
                    "ERROR",
                    status_fn=ToilStatus.getStatus,
                    process=wf,
                    seconds=60,
                )

    @needs_cwl
    @needs_docker
    @pytest.mark.cwl
    @pytest.mark.docker
    @pytest.mark.online
    def testGetStatusSuccessfulCWLWF(self, tmp_path: Path) -> None:
        """Test that ToilStatus.getStatus() behaves as expected with a successful CWL workflow."""
        jobstore = tmp_path / "jobstore"
        with get_data("test/cwl/sorttool.cwl") as cwl_file:
            with get_data("test/cwl/whale.txt") as input_file:
                cmd = [
                    "toil-cwl-runner",
                    "--jobStore",
                    str(jobstore),
                    "--clean=never",
                    str(cwl_file),
                    "--reverse",
                    "--input",
                    str(input_file),
                    f"--outdir={tmp_path}",
                ]
                wf = subprocess.Popen(cmd)
                self.check_status(
                    jobstore,
                    "RUNNING",
                    status_fn=ToilStatus.getStatus,
                    process=wf,
                    seconds=60,
                )
                wf.wait()
                self.check_status(
                    jobstore,
                    "COMPLETED",
                    status_fn=ToilStatus.getStatus,
                    process=wf,
                    seconds=60,
                )

    @needs_cwl
    @pytest.mark.cwl
    def testPrintJobLog(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test that ToilStatus.printJobLog() reads the log from a failed command without error."""
        jobstore = tmp_path / "jobstore"
        print_args: list[str] = []

        def fake_print(*args: Any, **kwargs: Any) -> None:
            print_args.extend(args)
        # Run a workflow that will always fail
        with get_data("test/cwl/alwaysfails.cwl") as cwl_file:
            cmd = [
                "toil-cwl-runner",
                "--logDebug",
                "--jobStore",
                str(jobstore),
                "--clean=never",
                f"--outdir={tmp_path/'outdir'}",
                str(cwl_file),
                "--message",
                "Testing",
            ]
            logger.info("Run command: %s", " ".join(cmd))
            wf = subprocess.Popen(cmd)
            wf.wait()
            # print log and check output
            status = ToilStatus(str(jobstore))
            with monkeypatch.context() as m:
                m.setattr(builtins, "print", fake_print)
                status.printJobLog()

        # Make sure it printed some kind of complaint about the missing command.
        assert "invalidcommand" in print_args[0]

    @pytest.mark.timeout(1200)
    def testRestartAttribute(self, tmp_path: Path) -> None:
        """
        Test that the job store is only destroyed when we observe a successful workflow run.
        The following simulates a failing workflow that attempts to resume without restart().
        In this case, the job store should not be destroyed until restart() is called.
        """
        jobstore = tmp_path / "jobstore"
        # Run a workflow that will always fail
        cmd = [
            python,
            "-m",
            "toil.test.sort.restart_sort",
            jobstore.as_uri(),
            "--badWorker=1",
            "--logDebug",
        ]
        subprocess.run(cmd)

        restart_cmd = [
            python,
            "-m",
            "toil.test.sort.restart_sort",
            jobstore.as_uri(),
            "--badWorker=0",
            "--logDebug",
            "--restart",
        ]
        subprocess.run(restart_cmd)

        # Check the job store exists after restart attempt
        assert jobstore.exists()

        successful_cmd = [
            python,
            "-m",
            "toil.test.sort.sort",
            "--logDebug",
            jobstore.as_uri(),
            "--restart",
        ]
        subprocess.run(successful_cmd)

        # Check the job store is destroyed after calling restart()
        assert not jobstore.exists()


def printUnicodeCharacter() -> None:
    # We want to get a unicode character to stdout but we can't print it directly because of
    # Python encoding issues. To work around this we print in a separate Python process. See
    # http://stackoverflow.com/questions/492483/setting-the-correct-encoding-when-piping-stdout-in-python
    subprocess.check_call([sys.executable, "-c", "print('\\xc3\\xbc')"])


class RunTwoJobsPerWorker(Job):
    """
    Runs child job with same resources as self in an attempt to chain the jobs on the same worker
    """

    def run(self, fileStore: AbstractFileStore) -> None:
        self.addChildFn(printUnicodeCharacter)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os

from unittest import mock

from toil.lib.aws.ami import (
    aws_marketplace_flatcar_ami_search,
    feed_flatcar_ami_release,
    flatcar_release_feed_ami,
    get_flatcar_ami,
    ReleaseFeedUnavailableError
)
from toil.test import ToilTest, needs_aws_ec2, needs_online

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


@needs_online
class FlatcarFeedTest(ToilTest):
    """Test accessing the Flatcar AMI release feed, independent of the AWS API"""

    # Note that we need to support getting an empty list back, because
    # sometimes the Flatcar feeds are just down, and we can't fail CI at those
    # times.

    def test_parse_archive_feed(self):
        """Make sure we can get a Flatcar release from the Internet Archive."""
        ami = flatcar_release_feed_ami("us-west-2", "amd64", "archive")
        self.assertEqual(len(ami), len("ami-02b46c73fed689d1c"))
        self.assertTrue(ami.startswith("ami-"))

    def test_parse_beta_feed(self):
        """Make sure we can get a Flatcar release from the beta channel."""
        ami = flatcar_release_feed_ami("us-west-2", "amd64", "beta")
        self.assertEqual(len(ami), len("ami-02b46c73fed689d1c"))
        self.assertTrue(ami.startswith("ami-"))

    def test_parse_stable_feed(self):
        """Make sure we can get a Flatcar release from the stable channel."""
        ami = flatcar_release_feed_ami("us-west-2", "amd64", "stable")
        self.assertEqual(len(ami), len("ami-02b46c73fed689d1c"))
        self.assertTrue(ami.startswith("ami-"))


@needs_aws_ec2
class AMITest(ToilTest):
    @classmethod
    def setUpClass(cls):
        from toil.lib.aws.session import establish_boto3_session

        session = establish_boto3_session(region_name="us-west-2")
        cls.ec2_client = session.client("ec2")

    def test_fetch_flatcar(self):
        with self.subTest("Test flatcar AMI from user is prioritized."):
            with mock.patch.dict(os.environ, {"TOIL_AWS_AMI": "overridden"}):
                ami = get_flatcar_ami(self.ec2_client)
                self.assertEqual(ami, "overridden")

        with self.subTest("Test flatcar AMI returns an AMI-looking AMI."):
            try:
                ami = get_flatcar_ami(self.ec2_client)
                self.assertEqual(len(ami), len("ami-02b46c73fed689d1c"))
                self.assertTrue(ami.startswith("ami-"))
            except ReleaseFeedUnavailableError:
                # Ignore any remote systems being down.
                pass

        with self.subTest(
            "Test feed_flatcar_ami_release() returns an AMI-looking AMI."
        ):
            ami = feed_flatcar_ami_release(self.ec2_client, source="archive")
            self.assertTrue(ami is None or len(ami) == len("ami-02b46c73fed689d1c"))
            self.assertTrue(ami is None or ami.startswith("ami-"))

        with self.subTest(
            "Test aws_marketplace_flatcar_ami_search() returns an AMI-looking AMI."
        ):
            ami = aws_marketplace_flatcar_ami_search(self.ec2_client)
            self.assertTrue(ami is None or len(ami), len("ami-02b46c73fed689d1c"))
            self.assertTrue(ami is None or ami.startswith("ami-"))

    def test_fetch_arm_flatcar(self):
        """Test flatcar AMI finder architecture parameter."""
        try:
            ami = get_flatcar_ami(self.ec2_client, architecture="arm64")
            self.assertTrue(ami.startswith("ami-"))
        except ReleaseFeedUnavailableError:
            # Ignore any remote systems being down.
            pass



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import base64
import copy
import json
import logging
import os
import pickle
import random
import shutil
import signal
import socket
import stat
import subprocess
import sys
import threading
import time
import traceback
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any, Callable, Optional

from configargparse import ArgParser

from toil import logProcessContext
from toil.common import Config, Toil, safeUnpickleFromStream
from toil.cwl.utils import (
    CWL_UNSUPPORTED_REQUIREMENT_EXCEPTION,
    CWL_UNSUPPORTED_REQUIREMENT_EXIT_CODE,
)
from toil.deferred import DeferredFunctionManager
from toil.fileStores.abstractFileStore import AbstractFileStore
from toil.job import (
    CheckpointJobDescription,
    DebugStoppingPointReached,
    Job,
    JobDescription,
)
from toil.jobStores.abstractJobStore import AbstractJobStore
from toil.lib.expando import MagicExpando
from toil.lib.io import make_public_dir
from toil.lib.resources import ResourceMonitor
from toil.statsAndLogging import configure_root_logger, install_log_color, set_log_level

logger = logging.getLogger(__name__)


class StatsDict(MagicExpando):
    """Subclass of MagicExpando for type-checking purposes."""

    jobs: list[MagicExpando]

def nextChainable(
    predecessor: JobDescription, job_store: AbstractJobStore, config: Config
) -> Optional[JobDescription]:
    """
    Returns the next chainable job's JobDescription after the given predecessor
    JobDescription, if one exists, or None if the chain must terminate.

    :param predecessor: The job to chain from
    :param job_store: The JobStore to fetch JobDescriptions from.
    :param config: The configuration for the current run.
    """
    # If no more jobs to run or services not finished, quit
    if (
        predecessor.nextSuccessors() is None
        or len(predecessor.services) > 0
        or (
            isinstance(predecessor, CheckpointJobDescription)
            and predecessor.checkpoint is not None
        )
    ):
        logger.debug(
            "Stopping running chain of jobs: no successors: %s, services: %s, checkpoint: %s",
            predecessor.nextSuccessors() is None,
            len(predecessor.services),
            (
                isinstance(predecessor, CheckpointJobDescription)
                and predecessor.checkpoint is not None
            ),
        )
        return None

    # Get the next set of jobs to run
    jobs = list(predecessor.nextSuccessors() or set())
    if len(jobs) == 0:
        # If there are no jobs, we might just not have any children.
        logger.debug(
            "Stopping running chain of jobs because job has no ready children or follow-ons"
        )
        return None

    # If there are 2 or more jobs to run in parallel we quit
    if len(jobs) >= 2:
        logger.debug(
            "No more jobs can run in series by this worker," " it's got %i successors",
            len(jobs),
        )
        logger.debug("Two distinct successors are %s and %s", jobs[0], jobs[1])
        return None

    # Grab the only job that should be there.
    successorID = jobs[0]

    logger.debug("%s would chain to ID %s", predecessor, successorID)

    # Load the successor JobDescription
    successor = job_store.load_job(successorID)

    # We check the requirements of the successor to see if we can run it
    # within the current worker
    if successor.memory > predecessor.memory:
        logger.debug("We need more memory for the next job, so finishing")
        return None
    if successor.cores > predecessor.cores:
        logger.debug("We need more cores for the next job, so finishing")
        return None
    if successor.disk > predecessor.disk:
        logger.debug("We need more disk for the next job, so finishing")
        return None
    if successor.preemptible != predecessor.preemptible:
        logger.debug(
            "Preemptibility is different for the next job, returning to the leader"
        )
        return None
    if successor.predecessorNumber > 1:
        logger.debug(
            "The next job has multiple predecessors; we must return to the leader."
        )
        return None

    if len(successor.services) > 0:
        logger.debug(
            "The next job requires services that will not yet be started; we must return to the leader."
        )
        return None

    if isinstance(successor, CheckpointJobDescription):
        # Check if job is a checkpoint job and quit if so
        logger.debug("Next job is checkpoint, so finishing")
        return None

    if (
        not config.run_local_jobs_on_workers
        and predecessor.local
        and not successor.local
    ):
        # This job might be running on the leader, but the next job may not.
        #
        # TODO: Optimize by detecting whether we actually are on the leader,
        # somehow.
        logger.debug("Next job is not allowed to run on the leader, so finishing")
        return None

    # Made it through! This job is chainable.
    return successor

def unstick_worker(interval: float = 120, timeout: float = 120) -> None:
    """
    Thread function that tries to prevent the process from getting stuck.

    Meant to be used as a daemon thread: does not have a shutdown signal but
    cleans up on exit.

    :param interval: Try to unstick the process at intervals of this many
        seconds.
    :param timeout: Stop child processes that take longer than this many
        seconds to finish.
    """

    # We've observed Toil getting stuck reading the job from the job store,
    # either due to a problem with the FileJobStore or with local temp storage,
    # but then get unstuck as soon as someone logged in and ran lsof on the
    # Toil process. So we make sure to do that to ourselves every once in a
    # while as long as the worker is running.

    # Figure out our process ID
    pid = os.getpid()

    child: Optional[subprocess.Popen[bytes]] = None

    def clean_up_child() -> None:
        """
        Cleanup function to run at daemon thread shutdown when the main thread
        terminates without shutting us down.

        Also used to kill the child process if it takes too long.
        """
        if child is not None:
            # Kill the child immediately if it is running
            child.kill()
            try:
                # Wait one last time to try and reap the child process
                child.wait(timeout=5)
            except subprocess.TimeoutExpired:
                pass

    atexit.register(clean_up_child)

    # TODO: If we handle daemon thread shutdown just fine, why do we bother
    # with all the event stuff? Why not cut it?

    # Wait the interval before trying the first unstick
    time.sleep(interval)

    while True:
        # Run an lsof on our PID, which has been observed to unstick reads.
        #
        # We rely on the thread being able to go away and atexit() hooks
        # happening in the middle of a wait with a timeout.
        #
        # We also want to handle the case where the child process gets so
        # gummed up that it can't exit when killed.

        # Preserve errors form child process but not output
        child = subprocess.Popen(
            ["lsof", "-p", str(pid)],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
        try:
            child.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            logger.warning("Running lsof took too long!")
            clean_up_child()
            if child.returncode is None:
                # Kill didn't take
                logger.warning("Could not promptly kill child process: %s", child.pid)

        if child.returncode != 0:
            # Something went wrong, which is suspicious. Either it failed or it
            # timed out and could not be killed promptly.
            logger.warning("Could not list open files on ourselves. Return code: %s", child.returncode)

        # Wait the interval.
        time.sleep(interval)

def workerScript(
    job_store: AbstractJobStore,
    config: Config,
    job_name: str,
    job_store_id: str,
    redirect_output_to_log_file: bool = True,
    local_worker_temp_dir: Optional[str] = None,
    debug_flags: Optional[set[str]] = None,
) -> int:
    """
    Worker process script, runs a job.

    :param job_store: The JobStore to fetch JobDescriptions from.
    :param config: The configuration for the current run.
    :param job_name: The "job name" (a user friendly name) of the job to be run
    :param job_store_id: The job store ID of the job to be run
    :param redirect_output_to_log_file: If False, log directly to the console
        instead of capturing job output.
    :param local_worker_temp_dir: The directory for the worker to work in. May
        be recursively removed after the job runs.
    :param debug_flags: Flags to set on each job before running it.

    :return int: 1 if a job failed, or 0 if all jobs succeeded
    """

    configure_root_logger()
    set_log_level(config.logLevel)

    if config.colored_logs:
        install_log_color()

    logger.debug("Worker started for job %s...", job_name)

    ##########################################
    # Create the worker killer, if requested
    ##########################################

    logFileByteReportLimit = config.maxLogFileSize

    if config.badWorker > 0 and random.random() < config.badWorker:
        # We need to kill the process we are currently in, to simulate worker
        # failure. We don't want to just send SIGKILL, because we can't tell
        # that from a legitimate OOM on our CI runner. We're going to send
        # SIGUSR1 so our terminations are distinctive, and then SIGKILL if that
        # didn't stick. We definitely don't want to do this from *within* the
        # process we are trying to kill, so we fork off. TODO: We can still
        # leave the killing code running after the main Toil flow is done, but
        # since it's now in a process instead of a thread, the main Python
        # process won't wait around for its timeout to expire. I think this is
        # better than the old thread-based way where all of Toil would wait
        # around to be killed.

        killTarget = os.getpid()
        sleepTime = config.badWorkerFailInterval * random.random()
        if os.fork() == 0:
            # We are the child
            # Let the parent run some amount of time
            time.sleep(sleepTime)
            # Kill it gently
            os.kill(killTarget, signal.SIGUSR1)
            # Wait for that to stick
            time.sleep(0.01)
            try:
                # Kill it harder. Hope the PID hasn't already been reused.
                # If we succeeded the first time, this will OSError
                os.kill(killTarget, signal.SIGKILL)
            except OSError:
                pass
            # Exit without doing any of Toil's cleanup
            os._exit(0)

        # We don't need to reap the child. Either it kills us, or we finish
        # before it does. Either way, init will have to clean it up for us.

    ##########################################
    # Create the worker unsticker
    ##########################################
    unstick_thread = threading.Thread(target=unstick_worker, args=())
    unstick_thread.daemon = True
    unstick_thread.start()

    ##########################################
    # Load the environment for the job
    ##########################################

    # First load the environment for the job.
    with job_store.read_shared_file_stream("environment.pickle") as fileHandle:
        environment = safeUnpickleFromStream(fileHandle)
    env_reject = {
        "TMPDIR",
        "TMP",
        "HOSTNAME",
        "HOSTTYPE",
        "HOME",
        "LOGNAME",
        "USER",
        "DISPLAY",
        "JAVA_HOME",
        "XDG_SESSION_TYPE",
        "XDG_SESSION_CLASS",
        "XDG_SESSION_ID",
        "XDG_RUNTIME_DIR",
        "XDG_DATA_DIRS",
        "DBUS_SESSION_BUS_ADDRESS",
    }
    for i in environment:
        if i == "PATH":
            # Handle path specially. Sometimes e.g. leader may not include
            # /bin, but the Toil appliance needs it.
            if i in os.environ and os.environ[i] != "":
                # Use the provided PATH and then the local system's PATH
                os.environ[i] = environment[i] + ":" + os.environ[i]
            else:
                # Use the provided PATH only
                os.environ[i] = environment[i]
        elif i not in env_reject:
            os.environ[i] = environment[i]
    # sys.path is used by __import__ to find modules
    if "PYTHONPATH" in environment:
        for e in environment["PYTHONPATH"].split(":"):
            if e != "":
                sys.path.append(e)

    ##########################################
    # Setup the temporary directories.
    ##########################################
    # Dir to put all this worker's temp files in.
    if config.workflowID is None:
        raise RuntimeError("The worker workflow ID was never set.")
    toilWorkflowDir = Toil.getLocalWorkflowDir(config.workflowID, config.workDir)
    # Dir to put lock files in, ideally not on NFS.
    toil_coordination_dir = Toil.get_local_workflow_coordination_dir(
        config.workflowID, config.workDir, config.coordination_dir
    )
    if local_worker_temp_dir is None:
        # Invent a temp directory to work in
        local_worker_temp_dir = make_public_dir(toilWorkflowDir)
    os.chmod(local_worker_temp_dir, 0o755)

    ##########################################
    # Setup the logging
    ##########################################

    # This is mildly tricky because we don't just want to
    # redirect stdout and stderr for this Python process; we want to redirect it
    # for this process and all children. Consequently, we can't just replace
    # sys.stdout and sys.stderr; we need to mess with the underlying OS-level
    # file descriptors. See <http://stackoverflow.com/a/11632982/402891>

    # When we start, standard input is file descriptor 0, standard output is
    # file descriptor 1, and standard error is file descriptor 2.

    # Do we even want to redirect output? Let the config make us not do it.
    redirect_output_to_log_file = (
        redirect_output_to_log_file and not config.disableWorkerOutputCapture
    )

    # What file do we want to point FDs 1 and 2 to?
    tempWorkerLogPath = os.path.join(local_worker_temp_dir, "worker_log.txt")

    if redirect_output_to_log_file:
        # Announce that we are redirecting logging, and where it will now go.
        # This is only important if we are trying to manually trace a faulty worker invocation.
        logger.debug("Redirecting logging to %s", tempWorkerLogPath)
        sys.stdout.flush()
        sys.stderr.flush()

        # Save the original stdout and stderr (by opening new file descriptors
        # to the same files)
        origStdOut = os.dup(1)
        origStdErr = os.dup(2)

        # Open the file to send stdout/stderr to.
        logFh = os.open(tempWorkerLogPath, os.O_WRONLY | os.O_CREAT | os.O_APPEND)

        # Replace standard output with a descriptor for the log file
        os.dup2(logFh, 1)

        # Replace standard error with a descriptor for the log file
        os.dup2(logFh, 2)

        # Since we only opened the file once, all the descriptors duped from
        # the original will share offset information, and won't clobber each
        # others' writes. See <http://stackoverflow.com/a/5284108/402891>. This
        # shouldn't matter, since O_APPEND seeks to the end of the file before
        # every write, but maybe there's something odd going on...

        # Close the descriptor we used to open the file
        os.close(logFh)

    debugging = logging.getLogger().isEnabledFor(logging.DEBUG)
    ##########################################
    # Worker log file trapped from here on in
    ##########################################

    jobAttemptFailed = False
    failure_exit_code = 1
    first_job_cores = None
    statsDict = StatsDict()  # type: ignore[no-untyped-call]
    statsDict.jobs = []
    statsDict.workers.logs_to_leader = []
    statsDict.workers.logging_user_streams = []

    def blockFn() -> bool:
        return True

    job = None
    try:

        # Put a message at the top of the log, just to make sure it's working.
        logger.info("---TOIL WORKER OUTPUT LOG---")
        sys.stdout.flush()

        logProcessContext(config)

        ##########################################
        # Connect to the deferred function system
        ##########################################
        deferredFunctionManager = DeferredFunctionManager(toil_coordination_dir)

        ##########################################
        # Load the JobDescription
        ##########################################

        jobDesc = job_store.load_job(job_store_id)
        logger.debug("Parsed job description")

        ##########################################
        # Cleanup from any earlier invocation of the job
        ##########################################

        if not jobDesc.has_body():
            logger.debug("Job description has no body to run.")
            # Cleanup jobs already finished
            jobDesc.clear_nonexistent_dependents(job_store)
            logger.debug("Cleaned up any references to completed successor jobs")

        # This cleans the old log file which may
        # have been left if the job is being retried after a job failure.
        oldLogFile = jobDesc.logJobStoreFileID
        if oldLogFile is not None:
            jobDesc.logJobStoreFileID = None
            job_store.update_job(jobDesc)  # Update first, before deleting any files
            job_store.delete_file(oldLogFile)

        ##########################################
        # If a checkpoint exists, restart from the checkpoint
        ##########################################

        if (
            isinstance(jobDesc, CheckpointJobDescription)
            and jobDesc.checkpoint is not None
        ):
            # The job is a checkpoint, and is being restarted after previously completing
            logger.debug("Job is a checkpoint")
            # If the checkpoint still has extant successors or services, its
            # subtree didn't complete properly. We handle the restart of the
            # checkpoint here, removing its previous subtree.
            if next(jobDesc.successorsAndServiceHosts(), None) is not None:
                logger.debug("Checkpoint has failed; restoring")
                # Reduce the try count
                if jobDesc.remainingTryCount < 0:
                    raise RuntimeError("The try count of the job cannot be negative.")
                jobDesc.remainingTryCount = max(0, jobDesc.remainingTryCount - 1)
                jobDesc.restartCheckpoint(job_store)
            # Otherwise, the job and successors are done, and we can cleanup stuff we couldn't clean
            # because of the job being a checkpoint
            else:
                logger.debug(
                    "The checkpoint jobs seems to have completed okay, removing any checkpoint files to delete."
                )
                # Delete any remnant files
                list(
                    map(
                        job_store.delete_file,
                        list(
                            filter(
                                job_store.file_exists, jobDesc.checkpointFilesToDelete
                            )
                        ),
                    )
                )

        ##########################################
        # Setup the stats
        ##########################################

        # Remember the cores from the first job, which is how many we have reserved for us.
        statsDict.workers.requested_cores = jobDesc.cores
        startClock = ResourceMonitor.get_total_cpu_time()

        startTime = time.time()
        while True:
            ##########################################
            # Run the job body, if there is one
            ##########################################

            logger.info("Working on job %s", jobDesc)

            if jobDesc.has_body():
                # Load the job. It will use the same JobDescription we have been using.
                job = Job.loadJob(job_store, jobDesc)
                if isinstance(jobDesc, CheckpointJobDescription):
                    # If it is a checkpoint job, set the checkpoint
                    jobDesc.set_checkpoint()

                logger.info("Loaded body %s from description %s", job, jobDesc)

                if debug_flags:
                    for flag in debug_flags:
                        logger.debug("Turning on debug flag %s on job", flag)
                        job.set_debug_flag(flag)

                # Create a fileStore object for the job
                fileStore = AbstractFileStore.createFileStore(
                    job_store,
                    jobDesc,
                    local_worker_temp_dir,
                    blockFn,
                    caching=config.caching,
                )
                try:
                    with job._executor(
                        stats=statsDict, fileStore=fileStore
                    ):
                        with deferredFunctionManager.open() as defer:
                            with fileStore.open(job):
                                # Get the next block function to wait on committing this job
                                blockFn = fileStore.waitForCommit

                                # Run the job, save new successors, and set up
                                # locally (but don't commit) successor
                                # relationships and job completion.
                                # Pass everything as name=value because Cactus
                                # likes to override _runner when it shouldn't and
                                # it needs some hope of finding the arguments it
                                # wants across multiple Toil versions. We also
                                # still pass a jobGraph argument to placate old
                                # versions of Cactus.
                                job._runner(
                                    jobGraph=None,
                                    jobStore=job_store,
                                    fileStore=fileStore,
                                    defer=defer,
                                )

                                # When the executor for the job finishes it will
                                # kick off a commit with the link to the job body
                                # cut.
                finally:
                    # Accumulate messages from this job & any subsequent chained jobs.
                    # Keep the messages even if the job fails.
                    statsDict.workers.logs_to_leader += fileStore.logging_messages
                    statsDict.workers.logging_user_streams += (
                        fileStore.logging_user_streams
                    )

                logger.info("Completed body for %s", jobDesc)

            else:
                # The body may not be attached, in which case the
                # JobDescription is either a shell ready to be deleted or has
                # been scheduled after a failure to cleanup
                logger.debug("No user job to run, so finishing")
                break

            if AbstractFileStore._terminateEvent.is_set():
                raise RuntimeError("The termination flag is set")

            ##########################################
            # Establish if we can run another job within the worker
            ##########################################
            successor = nextChainable(jobDesc, job_store, config)
            if successor is None or config.disableChaining:
                # Can't chain any more jobs. We are going to stop.

                logger.info("Not chaining from job %s", jobDesc)

                # No need to commit because the _executor context manager did it.

                break

            logger.info("Chaining from %s to %s", jobDesc, successor)

            ##########################################
            # We have a single successor job that is not a checkpoint job. We
            # reassign the ID of the current JobDescription to the successor,
            # and take responsibility for both jobs' associated files in the
            # combined job.
            ##########################################

            # Make sure nothing has gone wrong and we can really chain
            if jobDesc.memory < successor.memory:
                raise RuntimeError(
                    "Cannot chain jobs. A job's memory cannot be less than it's successor."
                )
            if jobDesc.cores < successor.cores:
                raise RuntimeError(
                    "Cannot chain jobs. A job's cores cannot be less than it's successor."
                )

            # Save the successor's original ID, so we can clean it (and its
            # body) up after we finish executing it.
            successorID = successor.jobStoreID

            # Now we need to become that successor, under the original ID.
            successor.replace(jobDesc)
            jobDesc = successor

            # Clone the now-current JobDescription (which used to be the successor).
            # TODO: Why??? Can we not?
            jobDesc = copy.deepcopy(jobDesc)

            # Build a fileStore to update the job and commit the replacement.
            # TODO: can we have a commit operation without an entire FileStore???
            fileStore = AbstractFileStore.createFileStore(
                job_store,
                jobDesc,
                local_worker_temp_dir,
                blockFn,
                caching=config.caching,
            )

            # Update blockFn to wait for that commit operation.
            blockFn = fileStore.waitForCommit

            # This will update the job once the previous job is done updating
            fileStore.startCommit(jobState=True)

            logger.debug("Starting the next job")

        ##########################################
        # Finish up the stats
        ##########################################
        totalCPUTime, totalMemoryUsage = (
            ResourceMonitor.get_total_cpu_time_and_memory_usage()
        )
        statsDict.workers.time = str(time.time() - startTime)
        statsDict.workers.clock = str(totalCPUTime - startClock)
        statsDict.workers.memory = str(totalMemoryUsage)
        # Say the worker used the max disk we saw from any job
        max_bytes = 0
        for job_stats in statsDict.jobs:
            if "disk" in job_stats:
                max_bytes = max(max_bytes, int(job_stats.disk))
        statsDict.workers.disk = str(max_bytes)
        # Count the jobs executed.
        # TODO: toil stats could compute this but its parser is too general to hook into simply.
        statsDict.workers.jobs_run = len(statsDict.jobs)

        # log the worker log path here so that if the file is truncated the path can still be found
        if redirect_output_to_log_file:
            logger.info(
                "Worker log can be found at %s. Set --cleanWorkDir to retain this log",
                local_worker_temp_dir,
            )

        logger.info(
            "Finished running the chain of jobs on this node, we ran for a total of %f seconds",
            time.time() - startTime,
        )

    ##########################################
    # Trapping where worker goes wrong
    ##########################################
    except DebugStoppingPointReached:
        # Job wants the worker to stop for debugging
        raise
    except (
        BaseException
    ) as e:  # Case that something goes wrong in worker, or we are asked to stop
        if not isinstance(e, SystemExit):
            logger.critical(
                "Worker crashed with traceback:\n%s", traceback.format_exc()
            )
        logger.error(
            "Exiting the worker because of a failed job on host %s",
            socket.gethostname(),
        )
        if isinstance(e, CWL_UNSUPPORTED_REQUIREMENT_EXCEPTION):
            # We need to inform the leader that this is a CWL workflow problem
            # and it needs to inform its caller.
            failure_exit_code = CWL_UNSUPPORTED_REQUIREMENT_EXIT_CODE
        elif isinstance(e, SystemExit) and isinstance(e.code, int) and e.code != 0:
            # We're meant to be exiting with a particular code.
            failure_exit_code = e.code
        else:
            try:
                from WDL.runtime.error import CommandFailed

                if isinstance(e, CommandFailed):
                    failure_exit_code = e.exit_status
            except ImportError:
                # WDL dependency not available
                pass
        AbstractFileStore._terminateEvent.set()
    finally:
        # Get rid of our deferred function manager now so we can't mistake it
        # for someone else's if we do worker cleanup.
        del deferredFunctionManager
        try:
            import cwltool.main

            cwltool.main._terminate_processes()
        except (ImportError, ModuleNotFoundError):
            pass
        except Exception as e:
            logger.debug("cwltool.main._terminate_processess exception: %s", (e))
            raise e

    ##########################################
    # Wait for the asynchronous chain of writes/updates to finish
    ##########################################

    blockFn()

    ##########################################
    # All the asynchronous worker/update threads must be finished now,
    # so safe to test if they completed okay
    ##########################################

    if AbstractFileStore._terminateEvent.is_set():
        # Something has gone wrong.

        # Clobber any garbage state we have for this job from failing with
        # whatever good state is still stored in the JobStore
        jobDesc = job_store.load_job(job_store_id)
        # Remember that we failed
        jobAttemptFailed = True

    ##########################################
    # Cleanup
    ##########################################

    # Close the worker logging
    # Flush at the Python level
    sys.stdout.flush()
    sys.stderr.flush()
    if redirect_output_to_log_file:
        # Flush at the OS level
        os.fsync(1)
        os.fsync(2)

        # Close redirected stdout and replace with the original standard output.
        os.dup2(origStdOut, 1)

        # Close redirected stderr and replace with the original standard error.
        os.dup2(origStdErr, 2)

        # sys.stdout and sys.stderr don't need to be modified at all. We don't
        # need to call redirectLoggerStreamHandlers since they still log to
        # sys.stderr

        # Close our extra handles to the original standard output and standard
        # error streams, so we don't leak file handles.
        os.close(origStdOut)
        os.close(origStdErr)

    # Now our file handles are in exactly the state they were in before.

    # Copy back the log file to the global dir, if needed.
    # Note that we work with bytes instead of characters so we can seek
    # relative to the end (since Python won't decode Unicode backward, or even
    # interpret seek offsets in characters for us). TODO: We may get invalid or
    # just different Unicode by breaking up a character at the boundary!
    if jobAttemptFailed and redirect_output_to_log_file:
        jobDesc.logJobStoreFileID = logJobStoreFileID = job_store.getEmptyFileStoreID(
            jobDesc.jobStoreID, cleanup=True
        )
        with job_store.update_file_stream(logJobStoreFileID) as w:
            with open(tempWorkerLogPath, "rb") as f:
                if os.path.getsize(tempWorkerLogPath) > logFileByteReportLimit != 0:
                    if logFileByteReportLimit > 0:
                        f.seek(
                            -logFileByteReportLimit, 2
                        )  # seek to last tooBig bytes of file
                    elif logFileByteReportLimit < 0:
                        f.seek(
                            logFileByteReportLimit, 0
                        )  # seek to first tooBig bytes of file
                # Dump the possibly-invalid-Unicode bytes into the log file
                w.write(f.read())  # TODO load file using a buffer
        # Commit log file reference back to JobStore
        job_store.update_job(jobDesc)

    elif (
        debugging or (config.writeLogsFromAllJobs and not jobDesc.local)
    ) and redirect_output_to_log_file:  # write log messages
        with open(tempWorkerLogPath, "rb") as logFile:
            if os.path.getsize(tempWorkerLogPath) > logFileByteReportLimit != 0:
                if logFileByteReportLimit > 0:
                    logFile.seek(
                        -logFileByteReportLimit, 2
                    )  # seek to last tooBig bytes of file
                elif logFileByteReportLimit < 0:
                    logFile.seek(
                        logFileByteReportLimit, 0
                    )  # seek to first tooBig bytes of file
            # Make sure lines are Unicode so they can be JSON serialized as part of the dict.
            # We may have damaged the Unicode text by cutting it at an arbitrary byte so we drop bad characters.
            logMessages = [
                line.decode("utf-8", "skip") for line in logFile.read().splitlines()
            ]
        statsDict.logs.names = [names.stats_name for names in jobDesc.get_chain()]
        statsDict.logs.messages = logMessages

    # We have stats/logging to report back.
    # We report even if the job attempt failed.
    # TODO: Will that upset analysis of the stats?
    job_store.write_logs(json.dumps(statsDict, ensure_ascii=True))

    # Remove the temp dir
    cleanUp = config.cleanWorkDir
    if (
        cleanUp == "always"
        or (cleanUp == "onSuccess" and not jobAttemptFailed)
        or (cleanUp == "onError" and jobAttemptFailed)
    ):

        def make_parent_writable(func: Callable[[str], Any], path: str, _: Any) -> None:
            """
            When encountering an error removing a file or directory, make sure
            the parent directory is writable.

            cwltool likes to lock down directory permissions, and doesn't clean
            up after itself.
            """
            # Just chmod it for rwx for user. This can't work anyway if it isn't ours.
            try:
                os.chmod(
                    os.path.dirname(path), stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
                )
            except PermissionError as e:
                logger.error(
                    "Could not set permissions on %s to allow cleanup of %s: %s",
                    os.path.dirname(path),
                    path,
                    e,
                )

        shutil.rmtree(local_worker_temp_dir, onerror=make_parent_writable)

    # This must happen after the log file is done with, else there is no place to put the log
    if (not jobAttemptFailed) and jobDesc.is_subtree_done():
        for merged_in in jobDesc.get_chain():
            # We can now safely get rid of the JobDescription, and all jobs it chained up
            job_store.delete_job(merged_in.job_store_id)

    if jobAttemptFailed:
        return failure_exit_code
    else:
        return 0


def parse_args(args: list[str]) -> Any:
    """
    Parse command-line arguments to the worker.
    """

    # Drop the program name
    args = args[1:]

    # Make the parser
    parser = ArgParser()

    # Now add all the options to it

    # Base required job information
    parser.add_argument("jobName", type=str, help="Text name of the job being run")
    parser.add_argument(
        "jobStoreLocator",
        type=str,
        help="Information required to connect to the job store",
    )
    parser.add_argument(
        "jobStoreID", type=str, help="ID of the job within the job store"
    )

    # Additional worker abilities
    parser.add_argument(
        "--context",
        default=[],
        action="append",
        help="""Pickled, base64-encoded context manager(s) to run job inside of.
                Allows the Toil leader to pass setup and cleanup work provided by the
                batch system, in the form of pickled Python context manager objects,
                that the worker can then run before/after the job on the batch
                system's behalf.""",
    )

    return parser.parse_args(args)


@contextmanager
def in_contexts(contexts: list[str]) -> Iterator[None]:
    """
    Unpickle and enter all the pickled, base64-encoded context managers in the
    given list. Then do the body, then leave them all.
    """

    if len(contexts) == 0:
        # Base case: nothing to do
        yield
    else:
        first = contexts[0]
        rest = contexts[1:]

        try:
            manager = pickle.loads(base64.b64decode(first.encode("utf-8")))
        except:
            exc_info = sys.exc_info()
            logger.error(
                "Exception while unpickling context manager: ", exc_info=exc_info
            )
            raise

        with manager:
            # After entering this first context manager, do the rest.
            # It might set up stuff so we can decode later ones.
            with in_contexts(rest):
                yield


def main(argv: Optional[list[str]] = None) -> None:
    if argv is None:
        argv = sys.argv
    # Parse our command line
    options = parse_args(argv)

    ##########################################
    # Load the jobStore/config file
    ##########################################

    job_store = Toil.resumeJobStore(options.jobStoreLocator)
    config = job_store.config

    with in_contexts(options.context):
        # Call the worker
        exit_code = workerScript(job_store, config, options.jobName, options.jobStoreID)

    # Exit with its return value
    sys.exit(exit_code)
